from typing import Optional
from buildzr.dsl import (
    Workspace,
    Person,
    SoftwareSystem,
    Container,
    Component,
    SystemLandscapeView,
    SystemContextView,
    ContainerView,
    ComponentView,
)

def test_system_landscape_view() -> Optional[None]:

    w = Workspace('workspace', scope='landscape')\
        .contains(
            Person("User"),
            SoftwareSystem("System A"),
            SoftwareSystem("System B"),
        )\
        .where(lambda w: [
            w.person().user >> "Uses" >> w.software_system().system_a,
            w.software_system().system_a >> "Interacts with" >> w.software_system().system_b,
        ], implied=True)\
        .with_views(
            SystemLandscapeView(
                key="system_landscape_view_00",
                description="System Landscape View Test",
            ),
        )\
        .get_workspace()

    assert any(w.model.views.systemLandscapeViews)
    assert len(w.model.views.systemLandscapeViews) == 1

    user = w.person().user
    system_a = w.software_system().system_a
    system_b = w.software_system().system_b

    element_ids = list(map(lambda x: x.id, w.model.views.systemLandscapeViews[0].elements))
    relationship_ids = list(map(lambda x: x.id, w.model.views.systemLandscapeViews[0].relationships))

    assert len(element_ids) == 3
    assert {
        user.model.id,
        system_a.model.id,
        system_b.model.id,
    }.issubset(set(element_ids))

    assert len(relationship_ids) == 2
    assert {
        user.model.relationships[0].id,
        system_a.model.relationships[0].id,
    }.issubset(set(relationship_ids))

def test_system_context_view() -> Optional[None]:

    w = Workspace('w')\
            .contains(
                Person('u'),
                SoftwareSystem('email_system')\
                    .contains(
                        Container('email_c1'),
                        Container('email_c2'),
                    )\
                    .where(lambda s: [
                        s.email_c1 >> "Uses" >> s.email_c2,
                    ]),
                SoftwareSystem('business_app')
                    .contains(
                        Container('business_app_c1'),
                        Container('business_app_c2'),
                    )
                    .where(lambda s: [
                        s.business_app_c1 >> "Gets data from" >> s.business_app_c2,
                    ]),
                SoftwareSystem('git_repo'), # Unrelated!
                SoftwareSystem('external_system'), # Also unrelated!
            )\
            .where(lambda w: [
                w.person().u >> "Uses" >> w.software_system().business_app,
                w.person().u >> "Hacks" >> w.software_system().git_repo,
                w.software_system().business_app >> "Notifies users using" >> w.software_system().email_system,
                w.software_system().git_repo >> "Uses" >> w.software_system().external_system,
            ])\
            .with_views(
                SystemContextView(
                    software_system_selector=lambda w: w.software_system().business_app,
                    key="ss_business_app",
                    description="The business app",
                )
            )\
            .get_workspace()

    element_ids =  list(map(lambda x: x.id, w.model.views.systemContextViews[0].elements))
    relationship_ids =  list(map(lambda x: x.id, w.model.views.systemContextViews[0].relationships))

    print('element ids:', element_ids)
    print('person id:', w.person().u.model.id)
    print('email system id:', w.software_system().email_system.model.id)
    print('business app id:', w.software_system().business_app.model.id)
    print('git repo id:', w.software_system().git_repo.model.id)

    assert any(w.model.views.systemContextViews)
    assert len(w.model.views.systemContextViews) == 1
    assert len(element_ids) == 3
    assert len(relationship_ids) == 2
    assert w.person().u.model.id in element_ids
    assert w.software_system().business_app.model.id in element_ids
    assert w.software_system().email_system.model.id in element_ids
    assert w.software_system().git_repo.model.id not in element_ids
    assert w.software_system().business_app.business_app_c1.model.id not in element_ids
    assert w.software_system().business_app.business_app_c2.model.id not in element_ids
    assert w.software_system().email_system.email_c1.model.id not in element_ids
    assert w.software_system().email_system.email_c2.model.id not in element_ids
    assert w.software_system().business_app.model.relationships[0].id in relationship_ids
    assert w.software_system().business_app.model.relationships[0].sourceId == w.software_system().business_app.model.id
    assert w.software_system().business_app.model.relationships[0].destinationId == w.software_system().email_system.model.id
    assert w.person().u.model.relationships[0].id in relationship_ids
    assert w.person().u.model.relationships[0].sourceId == w.person().u.model.id
    assert w.person().u.model.relationships[0].destinationId == w.software_system().business_app.model.id

def test_system_context_view_with_exclude_user() -> Optional[None]:

    w = Workspace('w')\
            .contains(
                Person('u'),
                SoftwareSystem('email_system')\
                    .contains(
                        Container('email_c1'),
                        Container('email_c2'),
                    )\
                    .where(lambda s: [
                        s.email_c1 >> "Uses" >> s.email_c2,
                    ]),
                SoftwareSystem('business_app')
                    .contains(
                        Container('business_app_c1'),
                        Container('business_app_c2'),
                    )
                    .where(lambda s: [
                        s.business_app_c1 >> "Gets data from" >> s.business_app_c2,
                    ]),
                SoftwareSystem('git_repo'), # Unrelated!
                SoftwareSystem('external_system'), # Also unrelated!
            )\
            .where(lambda w: [
                w.person().u >> "Uses" >> w.software_system().business_app,
                w.person().u >> "Hacks" >> w.software_system().git_repo,
                w.software_system().business_app >> "Notifies users using" >> w.software_system().email_system,
                w.software_system().git_repo >> "Uses" >> w.software_system().external_system,
            ])\
            .with_views(
                SystemContextView(
                    software_system_selector=lambda w: w.software_system().business_app,
                    key="ss_business_app",
                    description="The business app",
                    exclude_elements=[
                        lambda w, e: e == w.person().u,
                    ]
                )
            )\
            .get_workspace()

    element_ids =  list(map(lambda x: x.id, w.model.views.systemContextViews[0].elements))
    relationship_ids =  list(map(lambda x: x.id, w.model.views.systemContextViews[0].relationships))

    print('element ids:', element_ids)
    print('person id:', w.person().u.model.id)
    print('email system id:', w.software_system().email_system.model.id)
    print('business app id:', w.software_system().business_app.model.id)
    print('git repo id:', w.software_system().git_repo.model.id)

    assert any(w.model.views.systemContextViews)
    assert len(w.model.views.systemContextViews) == 1
    assert len(element_ids) == 2
    assert len(relationship_ids) == 1
    assert w.person().u.model.id not in element_ids
    assert w.software_system().business_app.model.id in element_ids
    assert w.software_system().email_system.model.id in element_ids
    assert w.software_system().git_repo.model.id not in element_ids
    assert w.software_system().business_app.business_app_c1.model.id not in element_ids
    assert w.software_system().business_app.business_app_c2.model.id not in element_ids
    assert w.software_system().email_system.email_c1.model.id not in element_ids
    assert w.software_system().email_system.email_c2.model.id not in element_ids
    assert w.software_system().business_app.model.relationships[0].id in relationship_ids
    assert w.software_system().business_app.model.relationships[0].sourceId == w.software_system().business_app.model.id
    assert w.software_system().business_app.model.relationships[0].destinationId == w.software_system().email_system.model.id

    # We're excluding the user in the view. Its relationship with the software
    # system shouldn't be shown as well.
    assert w.person().u.model.relationships[0].id not in relationship_ids
    assert w.person().u.model.relationships[0].sourceId == w.person().u.model.id
    assert w.person().u.model.relationships[0].destinationId == w.software_system().business_app.model.id

def test_container_view() -> Optional[None]:

    w = Workspace('w')\
            .contains(
                Person('user'),
                SoftwareSystem('app')
                    .contains(
                        Container('web_application'),
                        Container('database'),
                    )
                    .where(lambda app: [
                        app.web_application >> "Reads from and writes to" >> app.database,
                    ]),
                SoftwareSystem('git_repo'), # Unrelated!
                SoftwareSystem('external_system'), # Also unrelated!
            )\
            .where(lambda w: [
                w.person().user >> "Uses" >> w.software_system().app.web_application,
                w.person().user >> "Hacks" >> w.software_system().git_repo,
                w.software_system().git_repo >> "Uses" >> w.software_system().external_system,
            ])\
            .with_views(
                ContainerView(
                    software_system_selector=lambda w: w.software_system().app,
                    key="ss_business_app",
                    description="The business app",
                )
            )\
            .get_workspace()

    element_ids =  list(map(lambda x: x.id, w.model.views.containerViews[0].elements))
    relationship_ids =  list(map(lambda x: x.id, w.model.views.containerViews[0].relationships))

    print('element ids:', element_ids)
    print('person id:', w.person().user.model.id)
    print('app id:', w.software_system().app.model.id)
    print('  web application id:', w.software_system().app.web_application.model.id)
    print('  database id:', w.software_system().app.database.model.id)
    print('git repo id:', w.software_system().git_repo.model.id)
    print('external system id:', w.software_system().external_system.model.id)

    assert any(w.model.views.containerViews)
    assert len(w.model.views.containerViews) == 1
    assert len(element_ids) == 3 # Only the two containers of the selected software system + the user
    assert len(relationship_ids) == 2 # Only the one relationship between the two containers + with the user
    assert w.person().user.model.id in element_ids
    assert w.software_system().app.model.id not in element_ids
    assert w.software_system().git_repo.model.id not in element_ids
    assert w.software_system().app.web_application.model.id in element_ids
    assert w.software_system().app.database.model.id in element_ids
    assert w.person().user.model.relationships[0].id in relationship_ids

def test_component_view() -> Optional[None]:

    w = Workspace('workspace')\
        .contains(
            Person('User'),
            SoftwareSystem("Software System")\
            .contains(
                Container("Web Application")\
                .contains(
                    Component("Component 1"),
                    Component("Component 2"),
                )\
                .where(lambda app: [
                    app.component_1 >> "Uses" >> app.component_2,
                ]),
                Container("Database"),
            )\
            .where(lambda s: [
                s.web_application.component_2 >> "Reads from and writes to" >> s.database,
            ]),
        )\
        .where(lambda w: [
            w.person().user >> "Uses" >> w.software_system().software_system.web_application.component_1,
        ])\
        .with_views(
            ComponentView(
                container_selector=lambda w: w.software_system().software_system.web_application,
                key="web_application_container_00",
                description="Component View Test",
            )
        )\
        .get_workspace()

    element_ids =  list(map(lambda x: x.id, w.model.views.componentViews[0].elements))
    relationship_ids =  list(map(lambda x: x.id, w.model.views.componentViews[0].relationships))

    print("user id:", w.person().user.model.id)
    print("software system id", w.software_system().software_system.model.id)
    print("  web application id", w.software_system().software_system.web_application.model.id)
    print("    component 1 id", w.software_system().software_system.web_application.component_1.model.id)
    print("    component 2 id", w.software_system().software_system.web_application.component_2.model.id)
    print("  database id", w.software_system().software_system.database.model.id)

    assert any(w.model.views.componentViews)
    assert len(w.model.views.componentViews) == 1

    assert w.person().user.model.id in element_ids
    assert w.software_system().software_system.web_application.component_1.model.id in element_ids
    assert w.software_system().software_system.web_application.component_2.model.id in element_ids
    assert w.software_system().software_system.database.model.id in element_ids

    assert w.person().user.model.relationships[0].id in relationship_ids
    assert w.person().user.model.relationships[0].sourceId in element_ids
    assert w.person().user.model.relationships[0].destinationId in element_ids
    assert w.software_system().software_system.web_application.component_1.model.relationships[0].id in relationship_ids
    assert w.software_system().software_system.web_application.component_2.model.relationships[0].id in relationship_ids
    assert w.software_system().software_system.web_application.component_2.model.relationships[0].sourceId in element_ids
    assert w.software_system().software_system.web_application.component_2.model.relationships[0].destinationId in element_ids

def test_component_view_with_exclude_user() -> Optional[None]:

    w = Workspace('workspace')\
        .contains(
            Person('User'),
            SoftwareSystem("Software System")\
            .contains(
                Container("Web Application")\
                .contains(
                    Component("Component 1"),
                    Component("Component 2"),
                )\
                .where(lambda app: [
                    app.component_1 >> "Uses" >> app.component_2,
                ]),
                Container("Database"),
            )\
            .where(lambda s: [
                s.web_application.component_2 >> "Reads from and writes to" >> s.database,
            ]),
        )\
        .where(lambda w: [
            w.person().user >> "Uses" >> w.software_system().software_system.web_application.component_1,
        ])\
        .with_views(
            ComponentView(
                container_selector=lambda w: w.software_system().software_system.web_application,
                key="web_application_container_00",
                description="Component View Test",
                exclude_elements=[
                    lambda w, e: e == w.person().user
                ]
            )
        )\
        .get_workspace()

    element_ids =  list(map(lambda x: x.id, w.model.views.componentViews[0].elements))
    relationship_ids =  list(map(lambda x: x.id, w.model.views.componentViews[0].relationships))

    print("user id:", w.person().user.model.id)
    print("software system id", w.software_system().software_system.model.id)
    print("  web application id", w.software_system().software_system.web_application.model.id)
    print("    component 1 id", w.software_system().software_system.web_application.component_1.model.id)
    print("    component 2 id", w.software_system().software_system.web_application.component_2.model.id)
    print("  database id", w.software_system().software_system.database.model.id)

    assert any(w.model.views.componentViews)
    assert len(w.model.views.componentViews) == 1

    assert w.person().user.model.id not in element_ids
    assert w.software_system().software_system.web_application.component_1.model.id in element_ids
    assert w.software_system().software_system.web_application.component_2.model.id in element_ids
    assert w.software_system().software_system.database.model.id in element_ids

    assert w.person().user.model.relationships[0].id not in relationship_ids
    assert w.person().user.model.relationships[0].sourceId not in element_ids
    assert w.person().user.model.relationships[0].destinationId in element_ids
    assert w.software_system().software_system.web_application.component_1.model.relationships[0].id in relationship_ids
    assert w.software_system().software_system.web_application.component_2.model.relationships[0].id in relationship_ids
    assert w.software_system().software_system.web_application.component_2.model.relationships[0].sourceId in element_ids
    assert w.software_system().software_system.web_application.component_2.model.relationships[0].destinationId in element_ids

def test_container_view_with_multiple_software_systems() -> Optional[None]:

    w = Workspace('workspace')\
        .contains(
            SoftwareSystem("App1")\
            .contains(
                Container("c1"),
            ),
            SoftwareSystem("App2")\
            .contains(
                Container("c2"),
            )
        )\
        .where(lambda w: [
            w.software_system().app1.c1 >> "uses" >> w.software_system().app2.c2,
        ])\
        .with_views(
            ContainerView(
                key="container_view_00",
                description="Container View Test",
                software_system_selector=lambda w: w.software_system().app1,
            ),
        )\
        .get_workspace()

    assert any(w.model.views.containerViews)
    assert len(w.model.views.containerViews) == 1

    app1 = w.software_system().app1
    app2 = w.software_system().app2
    c1 = app1.c1
    c2 = app2.c2

    element_ids =  list(map(lambda x: x.id, w.model.views.containerViews[0].elements))
    relationship_ids =  list(map(lambda x: x.id, w.model.views.containerViews[0].relationships))

    assert len(element_ids) == 2
    assert {
        c1.model.id,
        c2.model.id,
    }.issubset(set(element_ids))

    assert len(relationship_ids) == 1
    assert {
        c1.model.relationships[0].id,
    }.issubset(set(relationship_ids))


import pytest
from buildzr.dsl import (
    Workspace,
    SoftwareSystem,
    Person,
    Container,
    Component,
    expression,
    With,
)
from buildzr.dsl import Explorer
from typing import Optional, cast

@pytest.fixture
def workspace() -> Workspace:

    w = Workspace('w')\
        .contains(
            Person('u', tags={'user'}),
            SoftwareSystem('s', properties={
                'repo': 'https://github.com/amirulmenjeni/buildzr',
            })\
            .contains(
                Container('app'),
                Container('db', technology='mssql'),
            )\
            .where(lambda s: [
                s.app >> "Uses" >> s.db | With(tags={'backend-interface', 'mssql'}),
            ])
        )\
        .where(lambda w: [
            w.person().u >> "Uses" >> w.software_system().s | With(
                tags={'frontend-interface'},
                properties={
                    'url': 'http://example.com/docs/api/endpoint',
                }
            )
        ])

    return w

def test_filter_elements_by_tags(workspace: Workspace) -> Optional[None]:

    filter = expression.Expression(
        include_elements=[
            lambda w, e: 'Person' in e.tags,
            lambda w, e: 'Container' in e.tags,
            lambda w, e: 'user' in e.tags
        ]
    )

    elements = filter.elements(workspace)

    assert len(elements) == 3

def test_filter_elements_by_technology(workspace: Workspace) -> Optional[None]:

    # Note that some elements do not have technology attribute, like `Person` or
    # `SoftwareSystem`.
    #
    # This should not cause any problem to the filter.
    filter = expression.Expression(
        include_elements=[
            lambda w, e: e.technology == 'mssql',
        ]
    )

    elements = filter.elements(workspace)

    assert len(elements) == 1
    assert elements[0].model.name == 'db'

def test_filter_elements_by_sources_and_destinations(workspace: Workspace) -> Optional[None]:

    filter = expression.Expression(
        include_elements=[
            lambda w, e: 'u' in e.sources.names,
            lambda w, e: 'db' in e.destinations.names and 'Container' in e.destinations.tags,
        ]
    )

    elements = filter.elements(workspace)

    assert len(elements) == 2
    assert elements[0].model.name == 's'
    assert elements[1].model.name == 'app'

def test_filter_elements_by_properties(workspace: Workspace) -> Optional[None]:

    filter = expression.Expression(
        include_elements=[
            lambda w, e: 'repo' in e.properties.keys() and 'github.com' in e.properties['repo']
        ]
    )

    elements = filter.elements(workspace)

    assert len(elements) == 1
    assert elements[0].model.name == 's'

def test_filter_elements_by_equal_operator(workspace: Workspace) -> Optional[None]:

    filter = expression.Expression(
        include_elements=[
            lambda w, e: e == cast(SoftwareSystem, workspace.s).app,
        ]
    )

    elements = filter.elements(workspace)

    assert len(elements) == 1
    assert elements[0].model.name == 'app'

def test_include_all_elements(workspace: Workspace) -> Optional[None]:

    filter = expression.Expression()

    elements = filter.elements(workspace)

    all_elements = list(Explorer(workspace).walk_elements())

    assert len(elements) == len(all_elements)

def test_filter_relationships_by_tags(workspace: Workspace) -> Optional[None]:

    filter = expression.Expression(
        include_relationships=[
            lambda w, r: 'frontend-interface' in r.tags
        ]
    )

    elements = filter.elements(workspace)
    relationships = filter.relationships(workspace)
    all_elements = list(Explorer(workspace).walk_elements())

    assert len(relationships) == 1
    assert len(elements) == len(all_elements)
    assert relationships[0].source.model.name == 'u'
    assert relationships[0].destination.model.name == 's'

def test_filter_relationships_by_technology(workspace: Workspace) -> Optional[None]:

    filter = expression.Expression(
        include_relationships=[
            lambda w, r: 'mssql' in r.tags
        ]
    )

    elements = filter.elements(workspace)
    relationships = filter.relationships(workspace)
    all_elements = list(Explorer(workspace).walk_elements())

    assert len(relationships) == 1
    assert len(elements) == len(all_elements)
    assert relationships[0].source.model.name == 'app'
    assert relationships[0].destination.model.name == 'db'

def test_filter_relationships_by_source(workspace: Workspace) -> Optional[None]:

    filter = expression.Expression(
        include_relationships=[
            lambda w, r: r.source == cast(SoftwareSystem, workspace.s).app
        ]
    )

    elements = filter.elements(workspace)
    relationships = filter.relationships(workspace)
    all_elements = list(Explorer(workspace).walk_elements())

    assert len(relationships) == 1
    assert len(elements) == len(all_elements)
    assert relationships[0].source.model.name == 'app'
    assert relationships[0].destination.model.name == 'db'

def test_filter_relationships_by_destination(workspace: Workspace) -> Optional[None]:

    filter = expression.Expression(
        include_relationships=[
            lambda w, r: r.destination == cast(SoftwareSystem, workspace.s).db
        ]
    )

    elements = filter.elements(workspace)
    relationships = filter.relationships(workspace)
    all_elements = list(Explorer(workspace).walk_elements())

    assert len(relationships) == 1
    assert len(elements) == len(all_elements)
    assert relationships[0].source.model.name == 'app'
    assert relationships[0].destination.model.name == 'db'

def test_filter_relationships_by_properties(workspace: Workspace) -> Optional[None]:

    filter = expression.Expression(
        include_relationships=[
            lambda w, r: 'url' in r.properties.keys() and 'example.com' in r.properties['url']
        ]
    )

    elements = filter.elements(workspace)
    relationships = filter.relationships(workspace)
    all_elements = list(Explorer(workspace).walk_elements())

    assert len(relationships) == 1
    assert len(elements) == len(all_elements)
    assert 'url' in relationships[0].model.properties.keys()
    assert 'example.com' in relationships[0].model.properties['url']

def test_filter_element_with_workspace_path(workspace: Workspace) -> Optional[None]:

    filter = expression.Expression(
        include_elements=[
            lambda w, e: e == w.software_system().s.db,
        ]
    )

    elements = filter.elements(workspace)

    assert len(elements) == 1
    assert isinstance(elements[0], Container)
    assert elements[0].model.technology == 'mssql'

def test_filter_relationship_with_workspace_path(workspace: Workspace) -> Optional[None]:

    filter = expression.Expression(
        include_relationships=[
            lambda w, r: r.source == w.person().u
        ]
    )

    relationships = filter.relationships(workspace)

    assert len(relationships) == 1
    assert relationships[0].model.destinationId == workspace.software_system().s.model.id

def test_filter_elements_with_excludes(workspace: Workspace) -> Optional[None]:

    filter = expression.Expression(
        include_elements=[
            lambda w, e: 'Person' in e.tags,
            lambda w, e: 'Container' in e.tags,
            lambda w, e: 'user' in e.tags
        ],
        exclude_elements=[
            lambda w, e: e == w.person().u
        ]
    )

    elements = filter.elements(workspace)

    assert len(elements) == 2
    assert workspace.person().u.model.id not in list(map(lambda x: x.model.id, elements))

def test_filter_relationships_with_excludes(workspace: Workspace) -> Optional[None]:

    filter = expression.Expression(
        include_relationships=[
            lambda w, r: any(['interface' in tag for tag in r.tags]) # True for all relationships
        ],
        exclude_relationships=[
            lambda w, r: r.source == w.person().u
        ]
    )

    relationships = filter.relationships(workspace)
    assert len(relationships) == 1

def test_filter_elements_without_includes_only_excludes(workspace: Workspace) -> Optional[None]:

    filter = expression.Expression(
        exclude_elements=[
            lambda w, e: w.person().u == e
        ]
    )

    elements = filter.elements(workspace)
    assert workspace.person().u.model.id not in list(map(lambda x: x.model.id, elements))

def test_filter_relationships_without_includes_only_excludes(workspace: Workspace) -> Optional[None]:

    filter = expression.Expression(
        exclude_relationships=[
            lambda w, r: r.source == w.person().u
        ]
    )

    relationships = filter.relationships(workspace)
    assert len(relationships) == 1

def test_filter_type(workspace: Workspace) -> Optional[None]:
    # Create an expression with include_elements and exclude_elements

    filter = expression.Expression(
        include_elements=[
            lambda w, e: e.type == Person,
            lambda w, e: e.type == Container,
        ],
    )

    elements = filter.elements(workspace)

    assert {
        workspace.person().u.model.id,
        workspace.software_system().s.app.model.id,
        workspace.software_system().s.db.model.id,
    }.issubset({ id for id in map(lambda x: x.model.id, elements) })
    assert len(elements) == 3


from __future__ import annotations
import dataclasses, json
import enum
import humps
from buildzr.dsl.interfaces import DslElement, DslWorkspaceElement
from typing import Union, List, TYPE_CHECKING, Type, Any
from typing_extensions import TypeGuard

if TYPE_CHECKING:
    from _typeshed import DataclassInstance
    JsonEncodable = Union[DslElement, DslWorkspaceElement, DataclassInstance, enum.Enum]
else:
    # Need this so that when we're not type checking with mypy, we're still on
    # the clear.
    JsonEncodable = Union[Any, None]

def _is_dataclass(obj: JsonEncodable) -> TypeGuard['DataclassInstance']:
    """
    Make mypy happy by ensuring that `obj` is indeed a `DataclassInstance`, and
    not merely its `Type[DataclassInstance]`.
    """

    return dataclasses.is_dataclass(obj) and not isinstance(obj, type)

def _remove_nones(d: Union[dict[str, Any], List[Any]]) -> Union[dict[str, Any], List[Any]]:

    """
    Remove the `null` valued JSON objects, as they're causing problem when read
    by Structurizr.

    I'm lazy. This code is stolen and modified from https://stackoverflow.com/a/66127889.
    """

    if isinstance(d, dict):
        for key, value in list(d.items()):
            if isinstance(value, (list, dict)):
                d[key] = _remove_nones(value)
            elif value is None or key is None:
                del d[key]
    elif isinstance(d, list):
        if isinstance(d, list):
            d = [_remove_nones(item) for item in d if item is not None]

    return d

class JsonEncoder(json.JSONEncoder):
    def default(self, obj: JsonEncodable) -> Union[str, list, dict]:
        # Handle the default encoder the nicely wrapped DSL elements.
        if isinstance(obj, DslElement) or isinstance(obj, DslWorkspaceElement):
            return humps.camelize(_remove_nones(dataclasses.asdict(obj.model)))

        # Handle the default encoder for those `dataclass`es models generated in
        # `buildzr.model`
        elif _is_dataclass(obj):
            return humps.camelize(_remove_nones(dataclasses.asdict(obj)))

        # Handle the enums
        elif isinstance(obj, enum.Enum):
            return str(obj.value)

        return super().default(obj) #type: ignore[no-any-return]


import buildzr
from buildzr.dsl import *
from typing import cast
from ..abstract_builder import AbstractBuilder

class GroupsSample(AbstractBuilder):

    def build(self) -> buildzr.models.Workspace:

        w = Workspace("w", scope=None)\
            .contains(
                Group(
                    "Company 1",
                    SoftwareSystem("A")\
                    .contains(
                        Container("a1"),
                        Container("a2"),
                    )
                ),
                Group(
                    "Company 2",
                    SoftwareSystem("B")\
                    .contains(
                        Container("b1"),
                        Container("b2")
                        .contains(
                            Component("c1"),
                        )
                    )
                ),
                SoftwareSystem("C"),
            )\
            .where(lambda w: [
                w.software_system().a >> "Uses" >> w.software_system().b,
                w.software_system().a.container().a1 >> "Uses" >> w.software_system().b.container().b1,
                w.software_system().a >> "Uses" >> w.software_system().c,
            ])\
            .with_views(
                SystemLandscapeView(
                    key='groups-sample',
                    description="Groups Sample"
                ),
                SystemContextView(
                    key='groups-sample-a',
                    software_system_selector=lambda w: cast(SoftwareSystem, w.a),
                    description="Groups Sample - Software System A"
                ),
                SystemContextView(
                    key='groups-sample-b',
                    software_system_selector=lambda w: cast(SoftwareSystem, w.b),
                    description="Groups Sample - Software System B"
                ),
                ContainerView(
                    key='groups-sample-b2',
                    software_system_selector=lambda w: w.software_system().b,
                    description="Groups Sample - Container B2"
                ),
            )\
            .get_workspace()

        return w.model


from abc import ABC, abstractmethod
from typing import (
    Any,
    Optional,
    Union,
    TypeVar,
    Generic,
    List,
    Set,
    Tuple,
    Callable,
    overload,
    Sequence,
    cast,
)
from typing_extensions import (
    Self
)
import buildzr

Model = Union[
    buildzr.models.Workspace,
    buildzr.models.Person,
    buildzr.models.SoftwareSystem,
    buildzr.models.Container,
    buildzr.models.Component,
]

TSrc = TypeVar('TSrc', bound='DslElement', contravariant=True)
TDst = TypeVar('TDst', bound='DslElement', contravariant=True)

TParent = TypeVar('TParent', bound=Union['DslWorkspaceElement', 'DslElement'], covariant=True)
TChild = TypeVar('TChild', bound='DslElement', contravariant=True)

class BindLeftLate(ABC, Generic[TDst]):

    @abstractmethod
    def set_source(self, source: Any) -> None:
        pass

    @abstractmethod
    def get_relationship(self) -> 'Optional[DslRelationship[Any, TDst]]':
        pass

class BindLeft(ABC, Generic[TSrc, TDst]):

    # Note: an abstraction of _UsesFrom

    @abstractmethod
    def __rshift__(self, destination: TDst) -> 'DslRelationship[TSrc, TDst]':
        pass

class BindRight(ABC, Generic[TSrc, TDst]):

    @overload
    @abstractmethod
    def __rshift__(self, description_and_technology: Tuple[str, str]) -> BindLeft[TSrc, TDst]:
        ...

    @overload
    @abstractmethod
    def __rshift__(self, description: str) -> BindLeft[TSrc, TDst]:
        ...

    @overload
    @abstractmethod
    def __rshift__(self, multiple_destinations: List[BindLeftLate[TDst]]) -> 'List[DslRelationship[TSrc, TDst]]':
        ...

    @abstractmethod
    def __rshift__(self, other: Union[str, Tuple[str, str], List[BindLeftLate[TDst]]]) -> Union[BindLeft[TSrc, TDst], 'List[DslRelationship[TSrc, TDst]]']:
        ...

class DslWorkspaceElement(ABC):

    @property
    @abstractmethod
    def model(self) -> buildzr.models.Workspace:
        pass

    @property
    @abstractmethod
    def parent(self) -> None:
        pass

    @property
    @abstractmethod
    def children(self) -> Optional[Sequence['DslElement']]:
        pass

    def __contains__(self, other: 'DslElement') -> bool:
        return self.model.id == other.parent.model.id

class DslElement(BindRight[TSrc, TDst]):
    """An abstract class used to label classes that are part of the buildzr DSL"""

    @property
    @abstractmethod
    def model(self) -> Model:
        """
        Returns the `dataclass` of the `DslElement` that follows Structurizr's
        JSON Schema (see https://github.com/structurizr/json)
        """
        pass

    @property
    @abstractmethod
    def parent(self) -> Union[None, DslWorkspaceElement, 'DslElement']:
        pass

    @property
    @abstractmethod
    def children(self) -> Union[None, Sequence['DslElement']]:
        pass

    @property
    @abstractmethod
    def sources(self) -> List['DslElement']:
        pass

    @property
    @abstractmethod
    def destinations(self) -> List['DslElement']:
        pass

    @property
    @abstractmethod
    def tags(self) -> Set[str]:
        pass

    def uses(
        self,
        other: 'DslElement',
        description: Optional[str]=None,
        technology: Optional[str]=None,
        tags: Set[str]=set()) -> 'DslRelationship[Self, DslElement]':
        pass

    def __contains__(self, other: 'DslElement') -> bool:
        return self.model.id == other.parent.model.id

class DslRelationship(ABC, Generic[TSrc, TDst]):
    """
    An abstract class specially used to label classes that are part of the
    relationship definer in the buildzr DSL
    """

    @property
    @abstractmethod
    def model(self) -> buildzr.models.Relationship:
        pass

    @property
    @abstractmethod
    def tags(self) -> Set[str]:
        pass

    @property
    @abstractmethod
    def source(self) -> DslElement:
        pass

    @property
    @abstractmethod
    def destination(self) -> DslElement:
        pass

    def __contains__(self, other: 'DslElement') -> bool:
        return self.source.model.id == other.model.id or self.destination.model.id == other.model.id

class DslFluentRelationship(ABC, Generic[TParent]):

    """
    The abstract class that defines the interface for the fluent relationship
    definition, where the one or more relationships between elements accessible
    from the `TParent` element.
    """

    @abstractmethod
    def where(
        self,
        func: Callable[
            [TParent],
            Sequence[
                Union[
                    DslRelationship,
                    Sequence[DslRelationship]
                ]
            ]
        ]) -> TParent:
        pass

    @abstractmethod
    def get(self) -> TParent:
        pass

class DslViewsElement(ABC):

    @property
    @abstractmethod
    def model(self) -> buildzr.models.Views:
        pass

    @property
    @abstractmethod
    def parent(self) -> DslWorkspaceElement:
        pass



from buildzr.encoders import JsonEncoder
from buildzr.models import Workspace
from abc import ABCMeta, abstractmethod
import json

class AbstractBuilder(metaclass=ABCMeta):

    @abstractmethod
    def build(self) -> Workspace:
        """Builds the `Workspace` using buildzr.

        This is an abstract method, and the deriving class is required to
        overwrite this method to provide the workspace build definition.
        Otherwise, `NotImplementedError` will be raised.
        """

        raise NotImplementedError()

    def writes_json_to(self, path: str) -> None:
        """Encode the `Workspace` as JSON and write to a file.

        This method requires the `build` method to be implemented.
        """

        workspace = self.build()

        with open(path, 'w', encoding='utf-8') as f:
            json.dump(workspace, f, ensure_ascii=False, indent=4, cls=JsonEncoder)


# Container view example shown in the cookbook (with a bit of modifications):
# https://docs.structurizr.com/dsl/cookbook/container-view/

import buildzr
from buildzr.dsl import *
from ..abstract_builder import AbstractBuilder

class SampleContainerView(AbstractBuilder):

    def build(self) -> buildzr.models.Workspace:

        w = Workspace('w', scope=None)\
                .contains(
                    Person('user'),
                    SoftwareSystem('app')
                        .contains(
                            Container('web_application'),
                            Container('database'),
                        )
                        .where(lambda app: [
                            app.web_application >> "Reads from and writes to" >> app.database,
                        ]),
                    SoftwareSystem('git_repo'), # Unrelated!
                    SoftwareSystem('external_system'), # Also unrelated!
                )\
                .where(lambda w: [
                    w.user >> "Uses" >> w.software_system().app.web_application,
                    w.user >> "Hacks" >> w.git_repo,
                    w.git_repo >> "Uses" >> w.external_system,
                ])\
                .with_views(
                    ContainerView(
                        software_system_selector=lambda w: w.software_system().app,
                        key="ss_business_app",
                        description="The business app",
                    )
                )\
                .get_workspace()

        return w.model


import buildzr
from buildzr.dsl import *
from ..abstract_builder import AbstractBuilder

class SampleComponentView(AbstractBuilder):

    """
    An example of a component view, as seen in
    https://docs.structurizr.com/dsl/cookbook/component-view/.
    """

    def build(self) -> buildzr.models.Workspace:

        w = Workspace('workspace')\
            .contains(
                Person('User'),
                SoftwareSystem("Software System")\
                .contains(
                    Container("Web Application")\
                    .contains(
                        Component("Component 1"),
                        Component("Component 2"),
                    )\
                    .where(lambda app: [
                        app.component_1 >> "Uses" >> app.component_2,
                    ]),
                    Container("Database"),
                )\
                .where(lambda s: [
                    s.web_application.component_2 >> "Reads from and writes to" >> s.database,
                ]),
            )\
            .where(lambda w: [
                w.person().user >> "Uses" >> w.software_system().software_system.web_application.component_1,
            ])\
            .with_views(
                ComponentView(
                    container_selector=lambda w: w.software_system().software_system.web_application,
                    key="web_application_container_00",
                    description="Component View Test",
                )
            )\
            .get_workspace()

        return w.model


import buildzr
from buildzr.dsl import (
    Workspace,
    SoftwareSystem,
    Person,
    Container,
    SystemContextView,
)
from ..abstract_builder import AbstractBuilder

class SystemContextViewSample(AbstractBuilder):

    def build(self) -> buildzr.models.Workspace:

        w = Workspace('w')\
                .contains(
                    Person('user'),
                    SoftwareSystem('web_app')
                        .contains(
                            Container('database'),
                            Container('api'),
                        ),
                    SoftwareSystem('email_system'),
                )\
                .where(lambda w: [
                    w.person().user >> "uses" >> w.software_system().web_app,
                    w.software_system().web_app >> "sends notification using" >> w.software_system().email_system,
                ])\
                .with_views(
                    SystemContextView(
                        lambda w: w.software_system().web_app,
                        key='web_app_system_context_00',
                        description="Web App System Context",
                        exclude_elements=[
                            lambda w, e: w.person().user == e,
                        ]
                    )
                )\
                .get_workspace()

        return w.model


import pytest
from typing import Optional, cast
from buildzr.dsl import (
    Workspace,
    SoftwareSystem,
    Person,
    Container,
    Component,
    With,
)
from buildzr.dsl.interfaces import DslRelationship
from buildzr.dsl import Explorer

@pytest.fixture
def workspace() -> Workspace:

    w = Workspace("w")\
            .contains(
                Person("u"),
                SoftwareSystem("s")\
                    .contains(
                        Container("webapp")\
                            .contains(
                                Component("database layer"),
                                Component("API layer"),
                                Component("UI layer"),
                            )\
                            .where(lambda app: [
                                app.ui_layer >> ("Calls HTTP API from", "http/api") >> app.api_layer,
                                app.api_layer >> ("Runs queries from", "sql/sqlite") >> app.database_layer
                            ]),\
                        Container("database"),
                    )\
                    .where(lambda s: [
                        s.webapp >> "Uses" >> s.database
                    ])
            )\
            .where(lambda w: [
                w.person().u >> "Runs SQL queries" >> w.software_system().s.database
            ], implied=True)

    return w

def test_walk_elements(workspace: Workspace) -> Optional[None]:

    explorer = Explorer(workspace).walk_elements()
    assert next(explorer).model.name == 'u'
    assert next(explorer).model.name == 's'
    assert next(explorer).model.name == 'webapp'
    assert next(explorer).model.name == 'database layer'
    assert next(explorer).model.name == 'API layer'
    assert next(explorer).model.name == 'UI layer'
    assert next(explorer).model.name == 'database'

def test_walk_relationships(workspace: Workspace) -> Optional[None]:

    explorer = Explorer(workspace).walk_relationships()

    next_relationship = next(explorer)
    assert next_relationship.model.description == "Runs SQL queries"
    assert next_relationship.model.sourceId == cast(Person, workspace.u).model.id
    assert next_relationship.model.destinationId == cast(SoftwareSystem, workspace.s).database.model.id

    # Note: implied relationships
    next_relationship = next(explorer)
    assert next_relationship.model.description == "Runs SQL queries"
    assert next_relationship.model.sourceId == cast(Person, workspace.u).model.id
    assert next_relationship.model.destinationId == cast(SoftwareSystem, workspace.s).model.id

    next_relationship = next(explorer)
    assert next_relationship.model.description == "Uses"
    assert next_relationship.model.sourceId == cast(SoftwareSystem, workspace.s).webapp.model.id
    assert next_relationship.model.destinationId == cast(SoftwareSystem, workspace.s).database.model.id

    next_relationship = next(explorer)
    assert next_relationship.model.description == "Runs queries from"
    assert next_relationship.model.sourceId == cast(SoftwareSystem, workspace.s).webapp.api_layer.model.id
    assert next_relationship.model.destinationId == cast(SoftwareSystem, workspace.s).webapp.database_layer.model.id

    next_relationship = next(explorer)
    assert next_relationship.model.description == "Calls HTTP API from"
    assert next_relationship.model.sourceId == cast(SoftwareSystem, workspace.s).webapp.ui_layer.model.id
    assert next_relationship.model.destinationId == cast(SoftwareSystem, workspace.s).webapp.api_layer.model.id


from typing import (
    Any,
    List,
    Tuple,
    Union,
    Set,
    Generic,
    Optional,
    Callable,
    Sequence,
    Dict,
    cast,
    overload,
)
from typing_extensions import Self, TypeIs
from dataclasses import dataclass
from buildzr.dsl.interfaces import (
    BindLeft,
    BindLeftLate,
    DslRelationship,
    DslFluentRelationship,
    DslElement,
    DslWorkspaceElement,
    TSrc, TDst,
    TParent, TChild,
)
from buildzr.dsl.factory import GenerateId
import buildzr

def _is_software_fluent_relationship(
    obj: '_FluentRelationship[Any]'
) -> TypeIs['_FluentRelationship[buildzr.dsl.SoftwareSystem]']:
    return isinstance(obj._parent, buildzr.dsl.SoftwareSystem)

def _is_container_fluent_relationship(
    obj: '_FluentRelationship[Any]'
) -> TypeIs['_FluentRelationship[buildzr.dsl.Container]']:
    return isinstance(obj._parent, buildzr.dsl.Container)

@dataclass
class With:
    tags: Optional[Set[str]] = None
    properties: Optional[Dict[str, str]] = None
    url: Optional[str] = None

@dataclass
class _UsesData(Generic[TSrc]):
    relationship: buildzr.models.Relationship
    source: TSrc

def desc(value: str, tech: Optional[str]=None) -> '_RelationshipDescription[DslElement]':
    if tech is None:
        return _RelationshipDescription(value)
    else:
        return _RelationshipDescription(value, tech)

class _UsesFrom(BindLeft[TSrc, TDst]):

    def __init__(self, source: TSrc, description: str="", technology: str="") -> None:
        self.uses_data = _UsesData(
            relationship=buildzr.models.Relationship(
                id=GenerateId.for_relationship(),
                description=description,
                technology=technology,
                sourceId=str(source.model.id),
            ),
            source=source,
        )

    def __rshift__(self, destination: TDst) -> '_Relationship[TSrc, TDst]':
        if isinstance(destination, DslWorkspaceElement):
            raise TypeError(f"Unsupported operand type for >>: '{type(self).__name__}' and {type(destination).__name__}")
        return _Relationship(self.uses_data, destination)

class _Relationship(DslRelationship[TSrc, TDst]):

    @property
    def model(self) -> buildzr.models.Relationship:
        return self._m

    @property
    def tags(self) -> Set[str]:
        return self._tags

    @property
    def source(self) -> DslElement:
        return self._src

    @property
    def destination(self) -> DslElement:
        return self._dst

    def __init__(
            self,
            uses_data: _UsesData[TSrc],
            destination: TDst,
            tags: Set[str]=set(),
            _include_in_model: bool=True,
        ) -> None:

        self._m = uses_data.relationship
        self._tags = {'Relationship'}.union(tags)
        self._src = uses_data.source
        self._dst = destination
        self.model.tags = ','.join(self._tags)

        uses_data.relationship.destinationId = str(destination.model.id)

        if not isinstance(uses_data.source.model, buildzr.models.Workspace):
            uses_data.source.destinations.append(self._dst)
            self._dst.sources.append(self._src)
            if _include_in_model:
                if uses_data.source.model.relationships:
                    uses_data.source.model.relationships.append(uses_data.relationship)
                else:
                    uses_data.source.model.relationships = [uses_data.relationship]

        # Used to pass the `_UsesData` object as reference to the `__or__`
        # operator overloading method.
        self._ref: Tuple[_UsesData] = (uses_data,)

    def __or__(self, _with: With) -> Self:
        return self.has(
            tags=_with.tags,
            properties=_with.properties,
            url=_with.url,
        )

    def has(
        self,
        tags: Optional[Set[str]]=None,
        properties: Optional[Dict[str, str]]=None,
        url: Optional[str]=None,
    ) -> Self:
        """
        Adds extra info to the relationship.

        This can also be achieved using the syntax sugar `DslRelationship |
        With(...)`.
        """
        if tags:
            self._tags = self._tags.union(tags)
            self._ref[0].relationship.tags = ",".join(self._tags)
        if properties:
            self._ref[0].relationship.properties = properties
        if url:
            self._ref[0].relationship.url = url
        return self

class _FluentRelationship(DslFluentRelationship[TParent]):

    """
    A hidden class used in the fluent DSL syntax after specifying a model (i.e.,
    Person, Software System, Container) to define relationship(s) within the
    specified model.
    """

    def __init__(self, parent: TParent) -> None:
        self._parent: TParent = parent

    def where(
        self,
        func: Callable[
            [TParent],
            Sequence[
                Union[
                    DslRelationship,
                    Sequence[DslRelationship]
                ]
            ]
        ], implied: bool=False) -> TParent:

        relationships: Sequence[DslRelationship] = []

        func = cast(Callable[[TParent], Sequence[Union[DslRelationship, Sequence[DslRelationship]]]], func)

        # Flatten the resulting relationship list.
        relationships = [
            rel for sublist in func(self._parent)
            for rel in (
                sublist if isinstance(sublist, list) else [sublist]
            )
        ]

        # If we have relationship s >> do >> a.b, then create s >> do >> a.
        # If we have relationship s.ss >> do >> a.b.c, then create s.ss >> do >> a.b and s.ss >> do >> a.
        # And so on...
        if implied:
            for relationship in relationships:
                source = relationship.source
                parent = relationship.destination.parent
                while parent is not None and not isinstance(parent, DslWorkspaceElement):
                    r = source.uses(parent, description=relationship.model.description, technology=relationship.model.technology)
                    r.model.linkedRelationshipId = relationship.model.id
                    parent = parent.parent

        return self._parent

    def get(self) -> TParent:
        return self._parent

class _RelationshipDescription(Generic[TDst]):

    def __init__(self, description: str, technology: Optional[str]=None) -> None:
        self._description = description
        self._technology = technology

    def __rshift__(self, destination: TDst) -> '_UsesFromLate[TDst]':
        if self._technology:
            return _UsesFromLate(
                description=(self._description, self._technology),
                destination=destination,
            )
        return _UsesFromLate(
            description=self._description,
            destination=destination,
        )

class _UsesFromLate(BindLeftLate[TDst]):
    """
    This method is used to create a relationship between one source element with
    multiple destination elements, like so:

    ```python
    u = Person("user")
    s1 = SoftwareSystem("software1")
    s2 = SoftwareSystem("software2")

    # Each element in the following list is a `_UsesFromLate` object.
    u >> [
        "Uses" >> s1 | With(tags={"linux", "rules"}),
        ("Reads from", "SQL") >> s1,
    ]
    ```

    This requires late left binding (i.e., the source element is bound after the
    the destination elements in the list are bounded. This is in contrast to how
    `_UsesFrom` works, where `u >> "Uses >> s1` binds the source element `u`
    first (i.e., in `u >> "Uses"` into `_UsesFrom` before finally binding `s1`
    in `((u >> "Uses) >> s2)`).
    """

    PossibleSourceType = Optional[DslElement]

    @dataclass
    class _LateBindData:
        tags: Optional[Set[str]] = None
        properties: Optional[Dict[str, str]] = None
        url: Optional[str] = None

    def __init__(self, description: Union[str, Tuple[str, str]], destination: TDst) -> None:
        if isinstance(description, str):
            self._description = description
            self._technology: Optional[str] = None
        elif isinstance(description, tuple) and len(description) == 2:
            self._description = description[0]
            self._technology = description[1]
        self._source: Optional[_UsesFromLate.PossibleSourceType] = None
        self._destination = destination
        self._relationship: Optional[_Relationship[_UsesFromLate.PossibleSourceType, TDst]] = None
        self._late_bind_data: _UsesFromLate._LateBindData = _UsesFromLate._LateBindData()

    def set_source(self, source: PossibleSourceType) -> None:
        self._source = source
        self._relationship =  _Relationship(
            uses_data=_UsesData(
                relationship=buildzr.models.Relationship(
                    id=GenerateId.for_relationship(),
                    description=self._description,
                    technology=self._technology,
                    sourceId=str(self._source.model.id),
                ),
                source=self._source,
            ),
            destination=self._destination,
        )
        self._late_bind_with()

    def get_relationship(self) -> Optional['_Relationship[PossibleSourceType, TDst]']:
        return self._relationship

    def _late_bind_with(self) -> None:
        """
        Binds tags, properties, url to the relationship.
        Called once the relationship is set.
        """
        self._relationship = self._relationship.has(
            tags=self._late_bind_data.tags,
            properties=self._late_bind_data.properties,
            url=self._late_bind_data.url,
        )


    def __or__(self, other: With) -> Self:
        self._late_bind_data.tags = other.tags
        self._late_bind_data.properties = other.properties
        self._late_bind_data.url = other.url

        return self

class DslElementRelationOverrides(DslElement):

    """
    Base class meant to be derived from to override the `__rshift__` method to
    allow for the `>>` operator to be used to create relationships between
    elements.
    """

    @overload # type: ignore[override]
    def __rshift__(self, description_and_technology: Tuple[str, str]) -> _UsesFrom[Self, DslElement]:
        ...

    @overload
    def __rshift__(self, description: str) -> _UsesFrom[Self, DslElement]:
        ...

    @overload
    def __rshift__(self, _RelationshipDescription: _RelationshipDescription[DslElement]) -> _UsesFrom[Self, DslElement]:
        ...

    @overload
    def __rshift__(self, multiple_destinations: List[_UsesFromLate[DslElement]]) -> List[_Relationship[Self, DslElement]]:
        ...

    def __rshift__(
            self,
            other: Union[
                str,
                Tuple[str, str],
                _RelationshipDescription[DslElement],
                List[_UsesFromLate[DslElement]]
            ]) -> Union[_UsesFrom[Self, DslElement], List[_Relationship[Self, DslElement]]]:
        if isinstance(other, str):
            return _UsesFrom(self, other)
        elif isinstance(other, tuple):
            return _UsesFrom(self, description=other[0], technology=other[1])
        elif isinstance(other, _RelationshipDescription):
            return _UsesFrom(self, description=other._description, technology=other._technology)
        elif isinstance(other, list):
            relationships = []
            for dest in other:
                dest.set_source(self)
                relationships.append(dest.get_relationship())
            return cast(List[_Relationship[Self, DslElement]], relationships)
        else:
            raise TypeError(f"Unsupported operand type for >>: '{type(self).__name__}' and {type(other).__name__}")

    def uses(
        self,
        other: DslElement,
        description: Optional[str]=None,
        technology: Optional[str]=None,
        tags: Set[str]=set()) -> _Relationship[Self, DslElement]:

        source = self

        uses_from = _UsesFrom[Self, DslElement](
            source=source,
            description=description,
            technology=technology
        )

        return _Relationship(
            uses_from.uses_data,
            destination=other,
            tags=tags,
        )


from .dsl import (
    Workspace,
    SoftwareSystem,
    Person,
    Container,
    Component,
    Group,
    SystemLandscapeView,
    SystemContextView,
    ContainerView,
    ComponentView,
)
from .relations import (
    desc,
    With,
)
from .explorer import Explorer
from .expression import Expression


# A simple example as shown in https://docs.structurizr.com/dsl/example.

from buildzr.encoders import *
from buildzr.models import *
from ..abstract_builder import AbstractBuilder

class Simple(AbstractBuilder):

    def build(self) -> Workspace:

        workspace_config = WorkspaceConfiguration(
            scope=Scope.Landscape
        )

        u = Person(
            id=str(1),
            name="User"
        )

        ss = SoftwareSystem(
            id=str(2),
            name='Software System'
        )

        r0 = Relationship(
            id=str(3),
            description="Uses",
            sourceId=u.id,
            destinationId=ss.id
        )

        # Note that in `r0`, `u` is the source element.
        u.relationships = [r0]

        workspace = Workspace(
            id=0,
            name='engineering',
            description='engineering apps landscape',
            model=Model(
                people=[u],
                softwareSystems=[ss]
            ),
            configuration=workspace_config
        )

        return workspace


from buildzr.dsl.dsl import (
    Person,
    SoftwareSystem,
    Container,
    Component,
)

from buildzr.dsl.relations import (
    _Relationship,
    _UsesData,
)

from typing import (
    Union,
    Generator,
    Iterable,
)

from buildzr.dsl.dsl import (
    Workspace,
)

class Explorer:

    def __init__(self, workspace_or_element: Union[Workspace, Person, SoftwareSystem, Container, Component]):
        self._workspace_or_element = workspace_or_element

    def walk_elements(self) -> Generator[Union[Person, SoftwareSystem, Container, Component], None, None]:
        if self._workspace_or_element.children:
            for child in self._workspace_or_element.children:
                explorer = Explorer(child).walk_elements()
                yield child
                yield from explorer

    def walk_relationships(self) -> Generator[_Relationship, None, None]:
        import buildzr
        from buildzr.dsl.factory.gen_id import GenerateId

        if self._workspace_or_element.children:
            for child in self._workspace_or_element.children:
                # Relationships aren't materialized in the `Workspace` or in any
                # of the `DslElement`s. As such, we need to recreate the `_Relationship` objects
                # from the Structurizr model.

                if child.model.relationships and child.destinations:
                    for relationship, destination in zip(child.model.relationships, child.destinations):
                        fake_relationship = _Relationship(
                            _UsesData(
                                relationship=buildzr.models.Relationship(
                                    id=relationship.id,
                                    description=relationship.description,
                                    properties=relationship.properties,
                                    technology=relationship.technology,
                                    tags=relationship.tags,
                                    sourceId=relationship.sourceId,
                                ),
                                source=child,
                            ),
                            destination=destination,
                            _include_in_model=False,
                        )
                        fake_relationship._tags = set(relationship.tags.split(','))

                        yield fake_relationship

                explorer = Explorer(child).walk_relationships()
                yield from explorer


# Container view example shown in the cookbook:
# https://docs.structurizr.com/dsl/cookbook/container-view/

import buildzr
from buildzr.dsl import *
from ..abstract_builder import AbstractBuilder

class ViewSugar(AbstractBuilder):

    def build(self) -> buildzr.models.Workspace:

        w = Workspace("w")\
            .contains(
                Person("u"),
                SoftwareSystem("s"),
            )\
            .where(lambda w: [
                w.u >> "Uses" >> w.s
            ])

        print(w.model)

        return w.model




# A simple example as shown in https://docs.structurizr.com/dsl/example.

import buildzr
from buildzr.dsl import *
from ..abstract_builder import AbstractBuilder

class SimpleDsl(AbstractBuilder):

    def build(self) -> buildzr.models.Workspace:

        workspace = Workspace("My workspace")
        user = Person("A user")
        software_system = SoftwareSystem("A software system")

        workspace.contains(user, software_system)

        user >> ("Uses", "CLI") >> software_system | With(
            tags={"linux", "rules"}
        )

        return workspace.model


from dataclasses import dataclass, fields
import inspect
import pytest
import importlib
from typing import Optional, cast
from buildzr.dsl.interfaces import DslRelationship
from buildzr.dsl import (
    Workspace,
    Group,
    SoftwareSystem,
    Person,
    Container,
    Component,
    With,
    SystemContextView,
    desc,
)
from buildzr.encoders import JsonEncoder

@dataclass
class DslHolder:
    """A `dataclass` for us to hold the objects created using the DSL.

This helps by allowing us to create the workspace and other DSL objects in the
fixture once to be reused across multiple tests.
"""

    workspace: Workspace
    software_system: SoftwareSystem
    person: Person
    container: Container
    component: Component

@pytest.fixture
def dsl() -> DslHolder:

    workspace = Workspace("My Workspace", "A happy place")
    software_system = SoftwareSystem("My Software System")
    person = Person("Super user")
    container = Container("My container")
    component = Component("My component")

    return DslHolder(
        workspace=workspace,
        software_system=software_system,
        person=person,
        container=container,
        component=component,
    )

def test_docstrings(dsl: DslHolder) -> Optional[None]:
    """The docstrings of the DSL object should match the one in the Structurizr schema."""

    models_module = importlib.import_module('buildzr.models')

    classes = [cls for _, cls in inspect.getmembers(models_module, inspect.isclass)]

    class_docstring = {}
    for cls in classes:
        class_name = cls.__name__
        class_doc = cls.__doc__

        if class_doc is None or len(class_doc) == 0:
            class_docstring[class_name] = str()
        else:
            class_docstring[class_name] = class_doc

    for field in fields(dsl):
        dsl_obj = getattr(dsl, field.name)
        dsl_name = dsl_obj.__class__.__name__
        dsl_doc = dsl_obj.__class__.__doc__
        assert dsl_doc is not None
        assert dsl_doc.strip() == class_docstring[dsl_name].strip()

def test_element_ids(dsl: DslHolder) -> Optional[None]:

    assert dsl.workspace._m.id is not None
    assert dsl.person._m.id is not None
    assert dsl.software_system._m.id is not None
    assert dsl.container._m.id is not None
    assert dsl.component._m.id is not None

def test_workspace_has_configuration(dsl: DslHolder) -> Optional[None]:

    assert dsl.workspace._m.configuration is not None

def test_relationship_dsl(dsl: DslHolder) -> Optional[None]:

    dsl.person >> ("uses", "cli") >> dsl.software_system

    assert dsl.person._m.relationships is not None
    assert len(dsl.person._m.relationships) == 1
    assert dsl.person._m.relationships[0].id is not None
    assert dsl.person._m.relationships[0].sourceId == dsl.person.model.id
    assert dsl.person._m.relationships[0].destinationId == dsl.software_system.model.id
    assert dsl.person._m.relationships[0].description == "uses"
    assert dsl.person._m.relationships[0].technology == "cli"

def test_relationship_with_extra_info_using_with(dsl: DslHolder) -> Optional[None]:

    dsl.person >> ("uses", "cli") >> dsl.software_system | With(
        tags={"bash", "terminal"},
        properties={
            "authentication": "ssh",
        },
        url="http://example.com/info/relationship-user-uses-cli",
    )

    assert "bash" in dsl.person.model.relationships[0].tags
    assert "terminal" in dsl.person.model.relationships[0].tags
    assert "authentication" in dsl.person.model.relationships[0].properties.keys()
    assert "ssh" in dsl.person.model.relationships[0].properties['authentication']
    assert "http://example.com/info/relationship-user-uses-cli" == dsl.person.model.relationships[0].url

def test_relationship_with_extra_info_using_has(dsl: DslHolder) -> Optional[None]:

    (dsl.person >> ("uses", "cli") >> dsl.software_system).has(
        tags={"bash", "terminal"},
        properties={
            "authentication": "ssh",
        },
        url="http://example.com/info/relationship-user-uses-cli",
    )

    assert "bash" in dsl.person.model.relationships[0].tags
    assert "terminal" in dsl.person.model.relationships[0].tags
    assert "authentication" in dsl.person.model.relationships[0].properties.keys()
    assert "http://example.com/info/relationship-user-uses-cli" == dsl.person.model.relationships[0].url

def test_relationship_using_uses_method(dsl: DslHolder) -> Optional[None]:

    dsl.person\
        .uses(
            dsl.software_system,
            description="browses",
            technology="browser")\
        .has(
            tags={"webapp"},
            properties={
                "url": "http://link.example.page"
            }
        )

    assert any(dsl.person.model.relationships)
    assert any(dsl.person.model.relationships[0].tags)
    assert any(dsl.person.model.relationships[0].properties.keys())
    assert dsl.person.model.relationships[0].sourceId == dsl.person.model.id
    assert dsl.person.model.relationships[0].destinationId == dsl.software_system.model.id
    assert dsl.person.model.relationships[0].description == "browses"
    assert dsl.person.model.relationships[0].technology == "browser"
    assert set(dsl.person.model.relationships[0].tags.split(',')) == {'Relationship', 'webapp'}
    assert dsl.person.model.relationships[0].properties['url'] == "http://link.example.page"

def test_relationship_dont_work_with_workspace(dsl: DslHolder) -> Optional[None]:

    with pytest.raises(TypeError):
        dsl.workspace >> "uses" >> dsl.person #type: ignore[operator]

    with pytest.raises(TypeError):
        dsl.person >> "uses" >> dsl.workspace #type: ignore[operator]

    with pytest.raises(TypeError):
        dsl.workspace >> "uses" >> dsl.software_system #type: ignore[operator]

def test_workspace_model_inclusion_dsl(dsl: DslHolder) -> Optional[None]:

    dsl.workspace.contains(dsl.person, dsl.software_system)

    assert any(dsl.workspace._m.model.people)
    assert any(dsl.workspace._m.model.softwareSystems)

def test_parenting(dsl: DslHolder) -> Optional[None]:

    dsl.workspace.contains(dsl.person, dsl.software_system)
    dsl.software_system.contains(dsl.container)
    dsl.container.contains(dsl.component)

    assert dsl.person.parent.model.id == dsl.workspace.model.id
    assert dsl.software_system.parent.model.id == dsl.workspace.model.id
    assert dsl.container.parent.model.id == dsl.software_system.model.id
    assert dsl.component.parent.model.id == dsl.container.model.id

def test_making_children(dsl: DslHolder) -> Optional[None]:

    dsl.workspace.contains(dsl.person, dsl.software_system)
    dsl.software_system.contains(dsl.container)
    dsl.container.contains(dsl.component)

    assert dsl.workspace.children[0].model.id == dsl.person.model.id
    assert dsl.workspace.children[1].model.id == dsl.software_system.model.id
    assert dsl.software_system.children[0].model.id == dsl.container.model.id
    assert dsl.container.children[0].model.id == dsl.component.model.id

def test_accessing_child_elements(dsl: DslHolder) -> Optional[None]:

    w = Workspace("w")\
            .contains(
                Person("u"),
                SoftwareSystem("s")\
                    .contains(
                        Container("webapp")\
                            .contains(
                                Component("database layer"),
                                Component("API layer"),
                                Component("UI layer"),
                            )\
                            .where(lambda webapp: [
                                webapp.database_layer >> ("Calls HTTP API from", "http/api") >> webapp.api_layer,
                                webapp.api_layer >> ("Runs queries from", "sql/sqlite") >> webapp.ui_layer,
                            ]),\
                        Container("database"),
                    )\
                    .where(lambda s: [
                        s.webapp >> "Uses" >> s.database
                    ])
            )\
            .get()


    assert type(w.u) is Person
    assert type(w.s) is SoftwareSystem
    assert type(w.s.webapp) is Container
    assert type(w.s.database) is Container
    assert type(w.s.webapp.api_layer) is Component

    if isinstance(w['s'], SoftwareSystem):
        assert type(w['s']['webapp']['database layer']) is Component

def test_relationship_definition_commutativity() -> Optional[None]:

    from buildzr.encoders import JsonEncoder
    import jsondiff #type: ignore[import-untyped]
    import json

    # For now, we have to cheat a bit and manually edit each entity's ID so they
    # they're not identified as differences between the two workspaces. This is
    # because the current IDs are running numbers across the same class of
    # `DslElements`s.
    #
    # So, hashtag TODO.

    w1 = Workspace("w")
    w1.model.id = 1
    u1 = Person("u")
    u1.model.id = "2"
    s1 = SoftwareSystem("s")
    s1.model.id = "3"
    u1 >> "Uses" >> s1
    u1.model.relationships[0].id = "4"
    w1.contains(u1, s1)

    w2 = Workspace("w")
    w2.model.id = 1
    u2 = Person("u")
    u2.model.id = "2"
    s2 = SoftwareSystem("s")
    s2.model.id = "3"
    w2.contains(u2, s2)
    u2 >> "Uses" >> s2
    u2.model.relationships[0].id = "4"

    json_str_w1 = json.dumps(w1, cls=JsonEncoder)
    json_str_w2 = json.dumps(w2, cls=JsonEncoder)
    differences = jsondiff.diff(
        json_str_w1,
        json_str_w2,
    )

    assert not differences

def test_relationship_returns_correct_type(dsl: DslHolder) -> Optional[None]:

    dsl.workspace.contains(
        dsl.person,
        dsl.software_system,
    )

    relationship = dsl.person >> "Uses" >> dsl.software_system

    assert isinstance(relationship, DslRelationship)
    assert relationship.model.description == "Uses"
    assert relationship.model.destinationId == dsl.software_system.model.id

def test_fluent_workspace_definition() -> Optional[None]:

    w = Workspace("w")\
            .contains(
                Person("u"),
                SoftwareSystem("s")\
                    .contains(
                        Container("webapp")\
                            .contains(
                                Component("database layer"),
                                Component("API layer"),
                                Component("UI layer"),
                            )\
                            .where(lambda webapp: [
                                webapp.ui_layer >> ("Calls HTTP API from", "http/api") >> webapp.api_layer,
                                webapp.api_layer >> ("Runs queries from", "sql/sqlite") >> webapp.database_layer,
                            ]),\
                        Container("database"),
                    )\
                    .where(lambda s: [
                        s.webapp >> "Uses" >> s.database | With(
                            tags={
                                'api'
                            },
                            properties={
                                'url': 'https://example.com/api'
                            }
                        )
                    ])
            )\
            .where(lambda w: [
                w.person().u >> "Uses" >> w.person().s | With(
                    tags={"5g-network"},
                )
            ])

    assert any(w.model.model.people)
    assert any(w.model.model.people[0].relationships)
    assert any(w.model.model.softwareSystems)
    assert any(w.model.model.softwareSystems[0].containers)
    assert any(w.model.model.softwareSystems[0].containers[0].relationships)
    assert any(w.model.model.softwareSystems[0].containers[0].components)
    assert any(w.model.model.softwareSystems[0].containers[0].components[1].relationships)
    assert any(w.model.model.softwareSystems[0].containers[0].components[2].relationships)
    assert not w.model.model.softwareSystems[0].containers[0].components[0].relationships
    assert 'api' in w.model.model.softwareSystems[0].containers[0].relationships[0].tags.split(',')
    assert 'url' in w.model.model.softwareSystems[0].containers[0].relationships[0].properties.keys()
    assert 'example.com' in w.model.model.softwareSystems[0].containers[0].relationships[0].properties['url']

def test_fluent_workspace_definition_without_contains_where() -> Optional[None]:

    """
    Expected behavior: Workspace.contains(...) and SoftwareSystem.contains(...)
    methods should not need to expect explicit follow-up chain method .get() to
    get the instance of its child class.
    """

    w = Workspace('w')\
        .contains(
            Person('u'),
            SoftwareSystem('s')\
            .contains(
                Container('db'),
                Container('app')\
                .contains(
                    Component('api layer'),
                    Component('model layer'),
                    Component('view layer'),
                )
            )
        )\
        .where(lambda w: [
            w.person().u >> "Makes API calls" >> w.software_system().s.app.api_layer,
        ])

    assert isinstance(w.u, Person)
    assert isinstance(w.s, SoftwareSystem)
    assert isinstance(w.s.db, Container)
    assert isinstance(w.s.app, Container)
    assert isinstance(w.s.app.api_layer, Component)
    assert isinstance(w.s.app.model_layer, Component)
    assert isinstance(w.s.app.view_layer, Component)

    w.s.app.model_layer >> "Uses" >> w.s.db

    assert any(w.s.app.model_layer.model.relationships)
    assert w.s.app.model_layer.model.relationships[0].description == "Uses"
    assert w.s.app.model_layer.model.relationships[0].destinationId == w.s.db.model.id

    assert any(w.u.model.relationships)
    assert w.u.model.relationships[0].description == "Makes API calls"
    assert w.u.model.relationships[0].destinationId == w.s.app.api_layer.model.id

def test_implied_relationship() -> Optional[None]:
    """
    See: https://docs.structurizr.com/java/implied-relationships#createimpliedrelationshipsunlessanyrelationshipexistsstrategy
    """

    # _I think_ the behavior of the implied relationship can't really be tested
    # in the authoring tool, as it is handled by the rendering tool (e.g., the
    # Structurizr Lite or Structurizr On-Premise).
    #
    # But this test ensures that cross-layer relationship _doesn't_ create new
    # relationship. For example, u -> s.database doesn't explicitly create a u
    # -> s relationship in the workspace JSON.

    w = Workspace("w")\
            .contains(
                Person("u"),
                SoftwareSystem("s")\
                    .contains(
                        Container("webapp")\
                            .contains(
                                Component("database layer"),
                                Component("API layer"),
                                Component("UI layer"),
                            )\
                            .where(lambda webapp: [
                                webapp.ui_layer >> ("Calls HTTP API from", "http/api") >> webapp.api_layer,
                                webapp.api_layer >> ("Runs queries from", "sql/sqlite") >> webapp.database_layer,
                            ]),\
                        Container("database"),
                    )\
                    .where(lambda s: [
                        s.webapp >> "Uses" >> s.database
                    ], implied=True)
            )\
            .where(lambda w: [
                w.person().u >> "Runs SQL queries" >> w.software_system().s.database
            ], implied=True)

    assert isinstance(w.u, Person)
    assert isinstance(w.s, SoftwareSystem)
    assert len(w.u.model.relationships) == 2 # Should have u >> R >> s and u >> R >> s.database

    assert w.u.model.relationships[0].description == "Runs SQL queries"
    assert w.u.model.relationships[0].sourceId == w.u.model.id
    assert w.u.model.relationships[0].destinationId == w.s.database.model.id

    assert w.u.model.relationships[1].description == "Runs SQL queries"
    assert w.u.model.relationships[1].sourceId == w.u.model.relationships[0].sourceId
    assert w.u.model.relationships[1].destinationId == w.s.model.id
    assert w.u.model.relationships[1].linkedRelationshipId == w.u.model.relationships[0].id

def test_tags_on_elements() -> Optional[None]:

    u = Person('My User', tags={'admin'})
    ss = SoftwareSystem('My Software System', tags={'External', 'Cloud'})
    container = Container('The Container', tags={'Database'})
    component = Component('A Component', tags={'Views'})

    assert set(u.model.tags.split(',')) == {'Element', 'Person', 'admin'}
    assert u.tags == {'Element', 'Person', 'admin'}

    assert set(ss.model.tags.split(',')) == {'Element', 'Software System', 'External', 'Cloud'}
    assert ss.tags == {'Element', 'Software System', 'External', 'Cloud'}

    assert set(container.model.tags.split(',')) == {'Element', 'Container', 'Database'}
    assert container.tags == {'Element', 'Container', 'Database'}

    assert set(component.model.tags.split(',')) == {'Element', 'Component', 'Views'}
    assert component.tags == {'Element', 'Component', 'Views'}

def test_tags_on_relationship_using_uses() -> Optional[None]:

    u = Person('u')
    s = SoftwareSystem('s')
    r = u.uses(s, 'Uses', tags={'Human-Computer Interaction'})

    assert set(r.model.tags.split(',')) == {'Relationship', 'Human-Computer Interaction'}
    assert r.tags == {'Relationship', 'Human-Computer Interaction'}

def test_tags_on_relationship_using_with() -> Optional[None]:


    u = Person('u')
    s = SoftwareSystem('s')
    r = (u >> "Uses" >> s | With(tags={'Human-Computer Interaction'}))

    assert set(r.model.tags.split(',')) == {'Relationship', 'Human-Computer Interaction'}
    assert r.tags == {'Relationship', 'Human-Computer Interaction'}

def test_source_destinations_in_dsl_elements() -> Optional[None]:

    w = Workspace("w")\
            .contains(
                Person("u"),
                SoftwareSystem("s")\
                    .contains(
                        Container("webapp")\
                            .contains(
                                Component("database layer"),
                                Component("API layer"),
                                Component("UI layer"),
                            )\
                            .where(lambda webapp: [
                                webapp.ui_layer >> ("Calls HTTP API from", "http/api") >> webapp.api_layer,
                                webapp.api_layer >> ("Runs queries from", "sql/sqlite") >> webapp.database_layer,
                            ]),\
                        Container("database"),
                    )\
                    .where(lambda s: [
                        s.webapp >> "Uses" >> s.database
                    ])
            )\
            .where(lambda w: [
                w.person().u >> "Runs SQL queries" >> w.software_system().s.database
            ], implied=True)

    assert isinstance(w.u, Person)
    assert isinstance(w.s, SoftwareSystem)

    assert len(w.u.sources) == 0

    assert len(w.s.sources) == 1
    assert {w.u.model.id}.issubset({src.model.id for src in w.s.sources})

    assert len(w.u.destinations) == 2
    assert {w.s.model.id, w.s.database.model.id}.issubset({dst.model.id for dst in w.u.destinations})

    assert len(w.s.destinations) == 0

    assert len(w.s.webapp.sources) == 0

    assert len(w.s.database.sources) == 2
    assert {w.u.model.id, w.s.webapp.model.id}.issubset({dst.model.id for dst in w.s.database.sources})

def test_contains_operator() -> Optional[None]:

    from buildzr.dsl import Explorer

    w = Workspace("w")\
            .contains(
                Person("u"),
                SoftwareSystem("s")\
                    .contains(
                        Container("webapp")\
                            .contains(
                                Component("database layer"),
                                Component("API layer"),
                                Component("UI layer"),
                            )\
                            .where(lambda webapp: [
                                webapp.ui_layer >> ("Calls HTTP API from", "http/api") >> webapp.api_layer,
                                webapp.api_layer >> ("Runs queries from", "sql/sqlite") >> webapp.database_layer,
                            ]),\
                        Container("database"),
                    )\
                    .where(lambda s: [
                        s.webapp >> "Uses" >> s.database
                    ])
            )\
            .where(lambda w: [
                w.person().u >> "Runs SQL queries" >> w.software_system().s.database
            ], implied=True)

    assert isinstance(w.u, Person)
    assert isinstance(w.s, SoftwareSystem)
    assert w.u in w
    assert w.s in w
    assert w.s.webapp in w.s
    assert w.s.database in w.s
    assert w.s.webapp.database_layer in w.s.webapp
    assert w.s.webapp.api_layer in w.s.webapp
    assert w.s.webapp.ui_layer in w.s.webapp
    assert w.s.webapp.ui_layer not in w.s.database

    relationships = Explorer(w).walk_relationships()
    assert any([
        r for r in list(relationships)
        if w.u in r and w.s in r and r.model.description == "Runs SQL queries"
    ])

def test_accessing_typed_dynamic_attributes() -> Optional[None]:

    w = Workspace("w")\
            .contains(
                Person("u"),
                SoftwareSystem("s")\
                    .contains(
                        Container("webapp")\
                            .contains(
                                Component("database layer"),
                                Component("API layer"),
                                Component("UI layer"),
                            )\
                            .where(lambda webapp: [
                                webapp.ui_layer >> ("Calls HTTP API from", "http/api") >> webapp.api_layer,
                                webapp.api_layer >> ("Runs queries from", "sql/sqlite") >> webapp.database_layer,
                            ]),\
                        Container("database"),
                    )\
                    .where(lambda s: [
                        s.webapp >> "Uses" >> s.database
                    ])
            )\
            .where(lambda w: [
                w.person().u >> "Runs SQL queries" >> w.software_system().s.database
            ], implied=True)

    assert 'Person' in w.person().u.tags
    assert 'Software System' in w.software_system().s.tags
    assert 'Container' in w.software_system().s.container().webapp.tags
    assert 'Component' in w.software_system().s.container().webapp.component().ui_layer.tags

def test_dsl_where_with_workspace() -> Optional[None]:

    print("test: test_dsl_where_with_workspace")

    w = Workspace("w")\
        .contains(
            Person("User"),
            SoftwareSystem("Software")\
            .contains(
                Container("UI"),
                Container("Database")
            )\
            .where(lambda s: [
                s.ui >> "Reads from and writes to" >> s.database
            ])
        )\
        .where(lambda w: [
            w.person().user >> "Uses" >> w.software_system().software
        ])

    assert len(w.children) == 2
    assert w.software_system().software.ui.model.relationships[0].description == "Reads from and writes to"
    assert w.person().user.model.relationships[0].description == "Uses"

def test_one_source_to_many_destinations_relationships_for_person() -> Optional[None]:

    w = Workspace("w")

    person = Person("User")
    s1 = SoftwareSystem("Software 1")
    s2 = SoftwareSystem("Software 1")

    relationships = person >> [
        desc("Uses") >> s1,
        desc("Gets data", "SQL") >> s2
    ]

    assert len(relationships) == 2
    assert relationships[0].model.description == "Uses"

    assert relationships[1].model.description == "Gets data"
    assert relationships[1].model.technology == "SQL"

def test_one_source_to_many_destinations_relationships_in_where_method() -> Optional[None]:

    w = Workspace('w', scope='landscape')\
        .contains(
            Person('Personal Banking Customer'),
            Person('Customer Service Staff'),
            Person('Back Office Staff'),
            SoftwareSystem('ATM'),
            SoftwareSystem('Internet Banking System'),
            SoftwareSystem('Email System'),
            SoftwareSystem('Mainframe Banking System'),
        )\
        .where(lambda w: [
            w.person().personal_banking_customer >> [
                desc("Withdraws cash using") >> w.software_system().atm,
                desc("Views account balance, and makes payments using") >> w.software_system().internet_banking_system,
                desc("Ask questions to") >> w.person().customer_service_staff,
            ],
            w.person().customer_service_staff >> "Uses" >> w.software_system().mainframe_banking_system,
            w.person().back_office_staff >> "Uses" >> w.software_system().mainframe_banking_system,
            w.software_system().atm >> "Uses" >> w.software_system().mainframe_banking_system,
            w.software_system().email_system >> "Sends e-mail to" >> w.person().personal_banking_customer,
            w.software_system().internet_banking_system >> [
                desc("Gets account information from, and makes payments using") >> w.software_system().mainframe_banking_system,
                desc("Sends e-mail using") >> w.software_system().email_system,
            ],
        ])

    relationships = w.person().personal_banking_customer.model.relationships
    assert len(relationships) == 3
    assert relationships[0].description == "Withdraws cash using"
    assert relationships[0].destinationId == w.software_system().atm.model.id
    assert relationships[1].description == "Views account balance, and makes payments using"
    assert relationships[1].destinationId == w.software_system().internet_banking_system.model.id
    assert relationships[2].description == "Ask questions to"
    assert relationships[2].destinationId == w.person().customer_service_staff.model.id

def test_one_to_one_relationship_creation_with_desc() -> Optional[None]:

    w = Workspace("w")\
        .contains(
            Person("User"),

            SoftwareSystem("Software 1")\
            .contains(
                Container("Container 1"),
                Container("Container 2"),
            )\
            .where(lambda s: [
                s.container_1 >> desc("Uses", "HTTP") >> s.container_2
            ]),

            SoftwareSystem("Software 2")\
            .contains(
                Container("Container 3")\
                .contains(
                    Component("Component 1"),
                    Component("Component 2"),
                )\
                .where(lambda c: [
                    c.component_1 >> desc("Uses", "TCP") >> c.component_2
                ])
            )
        )\
        .where(lambda w: [
            w.person().user >> desc("Uses", "CLI") >> w.software_system().software_1,
            w.software_system().software_1 >> desc("Uses", "SSH") >> w.software_system().software_2,
        ])\

    assert w.person().user.model.relationships[0].description == "Uses"
    assert w.software_system().software_1.model.relationships[0].technology == "SSH"
    assert w.software_system().software_1.container().container_1.model.relationships[0].technology == "HTTP"
    assert w.software_system().software_2.container().container_3.component_1.model.relationships[0].technology == "TCP"

def test_one_to_many_relationship_with_tags() -> Optional[None]:

    w = Workspace("w")\
        .contains(
            Person("User"),
            SoftwareSystem("Software 1"),
            SoftwareSystem("Software 2"),
        )\
        .where(lambda w: [
            w.person().user >> [
                desc("Uses") >> w.software_system().software_1 | With(tags={"CLI"}),
                desc("Uses") >> w.software_system().software_2 | With(tags={"UI"}),
            ]
        ])

    relationships = w.person().user.model.relationships
    assert len(relationships) == 2
    assert set(relationships[0].tags.split(',')) == {"CLI", "Relationship"}
    assert set(relationships[1].tags.split(',')) == {"UI", "Relationship"}

def test_dynamic_attribute_access_with_labels() -> Optional[None]:

    w = Workspace("w")\
        .contains(
            Person("Long Long Name").labeled('u'),
            SoftwareSystem("Boring Software").labeled('b'),
            SoftwareSystem("Tedious Software").labeled('t')\
            .contains(
                Container('Web User Interface').labeled('web')\
                .contains(
                    Component('Database Layer').labeled('db'),
                    Component('API Layer').labeled('api'),
                    Component('UI Layer').labeled('ui'),
                ),
                Container('SQL Server Database').labeled('mssql'),
            )
        )\
        .where(lambda w: [
            w.person().u >> [
                desc("Uses", "CLI") >> w.software_system().boring_software,
                desc("Uses", "UI") >> w.software_system().t,
            ]
        ])

    assert w.person().u.model.name == "Long Long Name"
    assert w.software_system().b.model.name == "Boring Software"
    assert w.software_system().t.web.model.name == "Web User Interface"
    assert w.software_system().t.container().web.db.model.name == "Database Layer"
    assert w.software_system().t.container().web.api.model.name == "API Layer"
    assert w.software_system().t.container().web.ui.model.name == "UI Layer"
    assert w.software_system().t.container().mssql.model.name == "SQL Server Database"

def test_grouping() -> Optional[None]:

    w = Workspace("w")\
        .contains(
            Group(
                "Company 1",
                SoftwareSystem("A")\
                .contains(
                    Container("a1"),
                    Container("a2"),
                )
            ),
            Group(
                "Company 2",
                SoftwareSystem("B")\
                .contains(
                    Container("b1"),
                    Container("b2")
                    .contains(
                        Component("c1"),
                    )
                )
            ),
            SoftwareSystem("C"),
        )\
        .where(lambda w: [
            w.software_system().a >> "Uses" >> w.software_system().b,
            w.software_system().a.container().a1 >> "Uses" >> w.software_system().b.container().b1,
            w.software_system().a >> "Uses" >> w.software_system().c,
        ])

    a = w.software_system().a
    b = w.software_system().b

    assert a.model.group == "Company 1"
    assert b.model.group == "Company 2"
    assert a.container().a1.model.group == "Company 1"
    assert a.container().a2.model.group == "Company 1"
    assert b.container().b1.model.group == "Company 2"
    assert b.container().b2.model.group == "Company 2"
    assert a.model.relationships[0].destinationId == b.model.id
    assert a.container().a1.model.relationships[0].destinationId == b.container().b1.model.id
    assert b.container().b2.component().c1.model.group == "Company 2"
    assert a.model.relationships[1].destinationId == w.software_system().c.model.id



from dataclasses import dataclass
import buildzr
from .factory import GenerateId
from typing_extensions import (
    Self,
    TypeGuard,
    TypeIs,
)
from typing import (
    Any,
    Union,
    Tuple,
    List,
    Set,
    Dict,
    Optional,
    Generic,
    TypeVar,
    Protocol,
    Callable,
    Iterable,
    Literal,
    cast,
    overload,
    Sequence,
    Type,
)

from buildzr.dsl.interfaces import (
    DslWorkspaceElement,
    DslElement,
    DslViewsElement,
    TSrc, TDst,
    TParent, TChild,
)
from buildzr.dsl.relations import (
    _is_software_fluent_relationship,
    _is_container_fluent_relationship,
    _Relationship,
    _RelationshipDescription,
    _FluentRelationship,
    DslElementRelationOverrides,
)

def _child_name_transform(name: str) -> str:
    return name.lower().replace(' ', '_')

# TODO: Remove, not used.
# def _create_linked_relationship_from(
#     relationship: '_Relationship[TSrc, TDst]') -> buildzr.models.Relationship:

#     src = relationship.source
#     dst = relationship.destination

#     return buildzr.models.Relationship(
#         id=GenerateId.for_relationship(),
#         sourceId=str(src.model.id),
#         destinationId=str(dst.parent.model.id),
#         linkedRelationshipId=relationship.model.id,
#     )

TypedModel = TypeVar('TypedModel')
class TypedDynamicAttribute(Generic[TypedModel]):

    def __init__(self, dynamic_attributes: Dict[str, Any]) -> None:
        self._dynamic_attributes = dynamic_attributes

    def __getattr__(self, name: str) -> TypedModel:
        return cast(TypedModel, self._dynamic_attributes.get(name))

class Workspace(DslWorkspaceElement):
    """
    Represents a Structurizr workspace, which is a wrapper for a software architecture model, views, and documentation.
    """

    @property
    def model(self) -> buildzr.models.Workspace:
        return self._m

    @property
    def parent(self) -> None:
        return None

    @property
    def children(self) -> Optional[List[Union['Person', 'SoftwareSystem']]]:
        return self._children

    def __init__(self, name: str, description: str="", scope: Literal['landscape', 'software_system', None]='software_system') -> None:
        self._m = buildzr.models.Workspace()
        self._parent = None
        self._children: Optional[List[Union['Person', 'SoftwareSystem']]] = []
        self._dynamic_attrs: Dict[str, Union['Person', 'SoftwareSystem']] = {}
        self.model.id = GenerateId.for_workspace()
        self.model.name = name
        self.model.description = description
        self.model.model = buildzr.models.Model(
            people=[],
            softwareSystems=[],
            deploymentNodes=[],
        )

        scope_mapper: Dict[
            str,
            Literal[buildzr.models.Scope.Landscape, buildzr.models.Scope.SoftwareSystem, None]
        ] = {
            'landscape': buildzr.models.Scope.Landscape,
            'software_system': buildzr.models.Scope.SoftwareSystem,
            None: None
        }

        self.model.configuration = buildzr.models.WorkspaceConfiguration(
            scope=scope_mapper[scope],
        )

    def _contains_group(
        self,
        name: str,
        *models: Union[
            'Person',
            'SoftwareSystem',
            _FluentRelationship['SoftwareSystem'],
            _FluentRelationship['Container'],
        ]
    ) -> None:

        def recursive_group_name_assign(software_system: 'SoftwareSystem') -> None:
            software_system.model.group = name
            for container in software_system.children:
                container.model.group = name
                for component in container.children:
                    component.model.group = name

        for model in models:
            if isinstance(model, Person):
                model.model.group = name
            elif isinstance(model, SoftwareSystem):
                recursive_group_name_assign(model)
            elif _is_software_fluent_relationship(model):
                recursive_group_name_assign(model._parent)
            elif _is_container_fluent_relationship(model):
                recursive_group_name_assign(model._parent._parent)

        self.contains(*models)

    def contains(
        self,
        *models: Union[
            'Group',
            'Person',
            'SoftwareSystem',
            _FluentRelationship['SoftwareSystem'],
            _FluentRelationship['Container'],
        ]) -> _FluentRelationship['Workspace']:

        for model in models:
            if isinstance(model, Group):
                self._contains_group(model._name, *model._elements)
            elif isinstance(model, Person):
                self.add_element(model)
            elif isinstance(model, SoftwareSystem):
                self.add_element(model)
            elif _is_software_fluent_relationship(model):
                self.add_element(model._parent)
            elif _is_container_fluent_relationship(model):
                self.add_element(model._parent._parent)
        return _FluentRelationship['Workspace'](self)

    def person(self) -> TypedDynamicAttribute['Person']:
        return TypedDynamicAttribute['Person'](self._dynamic_attrs)

    def software_system(self) -> TypedDynamicAttribute['SoftwareSystem']:
        return TypedDynamicAttribute['SoftwareSystem'](self._dynamic_attrs)

    def add_element(self, element: Union['Person', 'SoftwareSystem']) -> None:
        if isinstance(element, Person):
            self._m.model.people.append(element._m)
            element._parent = self
            self._dynamic_attrs[_child_name_transform(element.model.name)] = element
            if element._label:
                self._dynamic_attrs[_child_name_transform(element._label)] = element
            self._children.append(element)
        elif isinstance(element, SoftwareSystem):
            self._m.model.softwareSystems.append(element._m)
            element._parent = self
            self._dynamic_attrs[_child_name_transform(element.model.name)] = element
            if element._label:
                self._dynamic_attrs[_child_name_transform(element._label)] = element
            self._children.append(element)
        else:
            raise ValueError('Invalid element type: Trying to add an element of type {} to a workspace.'.format(type(element)))

    def with_views(
        self,
        *views: Union[
            'SystemLandscapeView',
            'SystemContextView',
            'ContainerView',
            'ComponentView',
        ]
    ) -> 'Views':
        return Views(self).contains(*views)

    def __getattr__(self, name: str) -> Union['Person', 'SoftwareSystem']:
        return self._dynamic_attrs[name]

    def __getitem__(self, name: str) -> Union['Person', 'SoftwareSystem']:
        return self._dynamic_attrs[_child_name_transform(name)]

    def __dir__(self) -> Iterable[str]:
        return list(super().__dir__()) + list(self._dynamic_attrs.keys())

class SoftwareSystem(DslElementRelationOverrides):
    """
    A software system.
    """

    @property
    def model(self) -> buildzr.models.SoftwareSystem:
        return self._m

    @property
    def parent(self) -> Optional[Workspace]:
        return self._parent

    @property
    def children(self) -> Optional[List['Container']]:
        return self._children

    @property
    def sources(self) -> List[DslElement]:
        return self._sources

    @property
    def destinations(self) -> List[DslElement]:
        return self._destinations

    @property
    def tags(self) -> Set[str]:
        return self._tags

    def __init__(self, name: str, description: str="", tags: Set[str]=set(), properties: Dict[str, Any]=dict()) -> None:
        self._m = buildzr.models.SoftwareSystem()
        self._parent: Optional[Workspace] = None
        self._children: Optional[List['Container']] = []
        self._sources: List[DslElement] = []
        self._destinations: List[DslElement] = []
        self._tags = {'Element', 'Software System'}.union(tags)
        self._dynamic_attrs: Dict[str, 'Container'] = {}
        self._label: Optional[str] = None
        self.model.id = GenerateId.for_element()
        self.model.name = name
        self.model.description = description
        self.model.tags = ','.join(self._tags)
        self.model.properties = properties

    def contains(
        self,
        *containers: Union['Container', _FluentRelationship['Container']]
    ) -> _FluentRelationship['SoftwareSystem']:
        if not self.model.containers:
            self.model.containers = []

        for child in containers:
            if isinstance(child, Container):
                self.add_element(child)
            elif _is_container_fluent_relationship(child):
                self.add_element(child._parent)
        return _FluentRelationship['SoftwareSystem'](self)

    def labeled(self, label: str) -> 'SoftwareSystem':
        self._label = label
        return self

    def container(self) -> TypedDynamicAttribute['Container']:
        return TypedDynamicAttribute['Container'](self._dynamic_attrs)

    def add_element(self, element: 'Container') -> None:
        if isinstance(element, Container):
            self.model.containers.append(element.model)
            element._parent = self
            self._dynamic_attrs[_child_name_transform(element.model.name)] = element
            if element._label:
                self._dynamic_attrs[_child_name_transform(element._label)] = element
            self._children.append(element)
        else:
            raise ValueError('Invalid element type: Trying to add an element of type {} to a software system.'.format(type(element)))

    def __getattr__(self, name: str) -> 'Container':
        return self._dynamic_attrs[name]

    def __getitem__(self, name: str) -> 'Container':
        return self._dynamic_attrs[_child_name_transform(name)]

    def __dir__(self) -> Iterable[str]:
        return list(super().__dir__()) + list(self._dynamic_attrs.keys())

class Person(DslElementRelationOverrides):
    """
    A person who uses a software system.
    """

    @property
    def model(self) -> buildzr.models.Person:
        return self._m

    @property
    def parent(self) -> Optional[Workspace]:
        return self._parent

    @property
    def children(self) -> None:
        """
        The `Person` element does not have any children, and will always return
        `None`.
        """
        return None

    @property
    def sources(self) -> List[DslElement]:
        return self._sources

    @property
    def destinations(self) -> List[DslElement]:
        return self._destinations

    @property
    def tags(self) -> Set[str]:
        return self._tags

    def __init__(self, name: str, description: str="", tags: Set[str]=set(), properties: Dict[str, Any]=dict()) -> None:
        self._m = buildzr.models.Person()
        self._parent: Optional[Workspace] = None
        self._sources: List[DslElement] = []
        self._destinations: List[DslElement] = []
        self._tags = {'Element', 'Person'}.union(tags)
        self._label: Optional[str] = None
        self.model.id = GenerateId.for_element()
        self.model.name = name
        self.model.description = description
        self.model.relationships = []
        self.model.tags = ','.join(self._tags)
        self.model.properties = properties

    def labeled(self, label: str) -> 'Person':
        self._label = label
        return self

class Container(DslElementRelationOverrides):
    """
    A container (something that can execute code or host data).
    """

    @property
    def model(self) -> buildzr.models.Container:
        return self._m

    @property
    def parent(self) -> Optional[SoftwareSystem]:
        return self._parent

    @property
    def children(self) -> Optional[List['Component']]:
        return self._children

    @property
    def sources(self) -> List[DslElement]:
        return self._sources

    @property
    def destinations(self) -> List[DslElement]:
        return self._destinations

    @property
    def tags(self) -> Set[str]:
        return self._tags

    def contains(self, *components: 'Component') -> _FluentRelationship['Container']:
        if not self.model.components:
            self.model.components = []
        for component in components:
            self.add_element(component)
        return _FluentRelationship['Container'](self)

    def __init__(self, name: str, description: str="", technology: str="", tags: Set[str]=set(), properties: Dict[str, Any]=dict()) -> None:
        self._m = buildzr.models.Container()
        self._parent: Optional[SoftwareSystem] = None
        self._children: Optional[List['Component']] = []
        self._sources: List[DslElement] = []
        self._destinations: List[DslElement] = []
        self._tags = {'Element', 'Container'}.union(tags)
        self._dynamic_attrs: Dict[str, 'Component'] = {}
        self._label: Optional[str] = None
        self.model.id = GenerateId.for_element()
        self.model.name = name
        self.model.description = description
        self.model.relationships = []
        self.model.technology = technology
        self.model.tags = ','.join(self._tags)
        self.model.properties = properties

    def labeled(self, label: str) -> 'Container':
        self._label = label
        return self

    def component(self) -> TypedDynamicAttribute['Component']:
        return TypedDynamicAttribute['Component'](self._dynamic_attrs)

    def add_element(self, element: 'Component') -> None:
        if isinstance(element, Component):
            self.model.components.append(element.model)
            element._parent = self
            self._dynamic_attrs[_child_name_transform(element.model.name)] = element
            if element._label:
                self._dynamic_attrs[_child_name_transform(element._label)] = element
            self._children.append(element)
        else:
            raise ValueError('Invalid element type: Trying to add an element of type {} to a container.'.format(type(element)))

    def __getattr__(self, name: str) -> 'Component':
        return self._dynamic_attrs[name]

    def __getitem__(self, name: str) -> 'Component':
        return self._dynamic_attrs[_child_name_transform(name)]

    def __dir__(self) -> Iterable[str]:
        return list(super().__dir__()) + list(self._dynamic_attrs.keys())

class Component(DslElementRelationOverrides):
    """
    A component (a grouping of related functionality behind an interface that runs inside a container).
    """

    @property
    def model(self) -> buildzr.models.Component:
        return self._m

    @property
    def parent(self) -> Optional[Container]:
        return self._parent

    @property
    def children(self) -> None:
        return None

    @property
    def sources(self) -> List[DslElement]:
        return self._sources

    @property
    def destinations(self) -> List[DslElement]:
        return self._destinations

    @property
    def tags(self) -> Set[str]:
        return self._tags

    def __init__(self, name: str, description: str="", technology: str="", tags: Set[str]=set(), properties: Dict[str, Any]=dict()) -> None:
        self._m = buildzr.models.Component()
        self._parent: Optional[Container] = None
        self._sources: List[DslElement] = []
        self._destinations: List[DslElement] = []
        self._tags = {'Element', 'Component'}.union(tags)
        self._label: Optional[str] = None
        self.model.id = GenerateId.for_element()
        self.model.name = name
        self.model.description = description
        self.model.technology = technology
        self.model.relationships = []
        self.model.tags = ','.join(self._tags)
        self.model.properties = properties

    def labeled(self, label: str) -> 'Component':
        self._label = label
        return self

class Group:

    def __init__(
        self,
        name: str,
        *elements: Union[
            Person,
            SoftwareSystem,
            _FluentRelationship[SoftwareSystem],
            _FluentRelationship[Container],
    ]) -> None:
        self._name = name
        self._elements = elements

_RankDirection = Literal['tb', 'bt', 'lr', 'rl']

_AutoLayout = Optional[
    Union[
        _RankDirection,
        Tuple[_RankDirection, float],
        Tuple[_RankDirection, float, float]
    ]
]

def _auto_layout_to_model(auto_layout: _AutoLayout) -> buildzr.models.AutomaticLayout:
    """
    See: https://docs.structurizr.com/dsl/language#autolayout
    """

    model = buildzr.models.AutomaticLayout()

    def is_auto_layout_with_rank_separation(\
        auto_layout: _AutoLayout,
    ) -> TypeIs[Tuple[_RankDirection, float]]:
        if isinstance(auto_layout, tuple):
            return len(auto_layout) == 2 and \
                    type(auto_layout[0]) is _RankDirection and \
                    type(auto_layout[1]) is float
        return False

    def is_auto_layout_with_node_separation(\
        auto_layout: _AutoLayout,
    ) -> TypeIs[Tuple[_RankDirection, float, float]]:
        if isinstance(auto_layout, tuple) and len(auto_layout) == 3:
            return type(auto_layout[0]) is _RankDirection and \
                   all([type(x) is float for x in auto_layout[1:]])
        return False

    map_rank_direction: Dict[_RankDirection, buildzr.models.RankDirection] = {
        'lr': buildzr.models.RankDirection.LeftRight,
        'tb': buildzr.models.RankDirection.TopBottom,
        'rl': buildzr.models.RankDirection.RightLeft,
        'bt': buildzr.models.RankDirection.BottomTop,
    }

    if auto_layout is not None:
        if is_auto_layout_with_rank_separation(auto_layout):
            d, rs = cast(Tuple[_RankDirection, float], auto_layout)
            model.rankDirection = map_rank_direction[cast(_RankDirection, d)]
            model.rankSeparation = rs
        elif is_auto_layout_with_node_separation(auto_layout):
            d, rs, ns = cast(Tuple[_RankDirection, float, float], auto_layout)
            model.rankDirection = map_rank_direction[cast(_RankDirection, d)]
            model.rankSeparation = rs
            model.nodeSeparation = ns
        else:
            model.rankDirection = map_rank_direction[cast(_RankDirection, auto_layout)]

    if model.rankSeparation is None:
        model.rankSeparation = 300
    if model.nodeSeparation is None:
        model.nodeSeparation = 300
    if model.edgeSeparation is None:
        model.edgeSeparation = 0
    if model.implementation is None:
        model.implementation = buildzr.models.Implementation.Graphviz
    if model.vertices is None:
        model.vertices = False

    return model

class SystemLandscapeView:

    from buildzr.dsl.expression import Expression, Element, Relationship

    @property
    def model(self) -> buildzr.models.SystemLandscapeView:
        return self._m

    @property
    def parent(self) -> Optional['Views']:
        return self._parent

    def __init__(
        self,
        key: str,
        description: str,
        auto_layout: _AutoLayout='tb',
        title: Optional[str]=None,
        include_elements: List[Callable[[Workspace, Element], bool]]=[],
        exclude_elements: List[Callable[[Workspace, Element], bool]]=[],
        include_relationships: List[Callable[[Workspace, Relationship], bool]]=[],
        exclude_relationships: List[Callable[[Workspace, Relationship], bool]]=[],
        properties: Optional[Dict[str, str]]=None,
    ) -> None:
        self._m = buildzr.models.SystemLandscapeView()
        self._parent: Optional['Views'] = None

        self._m.key = key
        self._m.description = description

        self._m.automaticLayout = _auto_layout_to_model(auto_layout)
        self._m.title = title
        self._m.properties = properties

        self._include_elements = include_elements
        self._exclude_elements = exclude_elements
        self._include_relationships = include_relationships
        self._exclude_relationships = exclude_relationships

    def _on_added(self) -> None:

        from buildzr.dsl.expression import Expression, Element, Relationship
        from buildzr.models import ElementView, RelationshipView

        expression = Expression(
            include_elements=self._include_elements,
            exclude_elements=self._exclude_elements,
            include_relationships=self._include_relationships,
            exclude_relationships=self._exclude_relationships,
        )

        workspace = self._parent._parent

        include_view_elements_filter: List[Callable[[Workspace, Element], bool]] = [
            lambda w, e: e.type == Person,
            lambda w, e: e.type == SoftwareSystem
        ]

        exclude_view_elements_filter: List[Callable[[Workspace, Element], bool]] = [
            lambda w, e: e.type == Container,
            lambda w, e: e.type == Component,
        ]

        include_view_relationships_filter: List[Callable[[Workspace, Relationship], bool]] = [
            lambda w, r: r.source.type == Person,
            lambda w, r: r.source.type == SoftwareSystem,
            lambda w, r: r.destination.type == Person,
            lambda w, r: r.destination.type == SoftwareSystem,
        ]

        expression = Expression(
            include_elements=self._include_elements + include_view_elements_filter,
            exclude_elements=self._exclude_elements + exclude_view_elements_filter,
            include_relationships=self._include_relationships + include_view_relationships_filter,
            exclude_relationships=self._exclude_relationships,
        )

        element_ids = map(
            lambda x: str(x.model.id),
            expression.elements(workspace)
        )

        relationship_ids = map(
            lambda x: str(x.model.id),
            expression.relationships(workspace)
        )

        self._m.elements = []
        for element_id in element_ids:
            self._m.elements.append(ElementView(id=element_id))

        self._m.relationships = []
        for relationship_id in relationship_ids:
            self._m.relationships.append(RelationshipView(id=relationship_id))

class SystemContextView:

    """
    If no filter is applied, this view includes all elements that have a direct
    relationship with the selected `SoftwareSystem`.
    """

    from buildzr.dsl.expression import Expression, Element, Relationship

    @property
    def model(self) -> buildzr.models.SystemContextView:
        return self._m

    @property
    def parent(self) -> Optional['Views']:
        return self._parent

    def __init__(
        self,
        software_system_selector: Callable[[Workspace], SoftwareSystem],
        key: str,
        description: str,
        auto_layout: _AutoLayout='tb',
        title: Optional[str]=None,
        include_elements: List[Callable[[Workspace, Element], bool]]=[],
        exclude_elements: List[Callable[[Workspace, Element], bool]]=[],
        include_relationships: List[Callable[[Workspace, Relationship], bool]]=[],
        exclude_relationships: List[Callable[[Workspace, Relationship], bool]]=[],
        properties: Optional[Dict[str, str]]=None,
    ) -> None:
        self._m = buildzr.models.SystemContextView()
        self._parent: Optional['Views'] = None

        self._m.key = key
        self._m.description = description

        self._m.automaticLayout = _auto_layout_to_model(auto_layout)
        self._m.title = title
        self._m.properties = properties

        self._selector = software_system_selector
        self._include_elements = include_elements
        self._exclude_elements = exclude_elements
        self._include_relationships = include_relationships
        self._exclude_relationships = exclude_relationships

    def _on_added(self) -> None:

        from buildzr.dsl.expression import Expression, Element, Relationship
        from buildzr.models import ElementView, RelationshipView

        # TODO: Refactor below codes. Similar patterns may exists for other views.
        # Maybe make the views a subclass of some abstract `BaseView` class?

        software_system = self._selector(self._parent._parent)
        self._m.softwareSystemId = software_system.model.id
        view_elements_filter: List[Callable[[Workspace, Element], bool]] = [
            lambda w, e: e == software_system,
            lambda w, e: software_system.model.id in e.sources.ids,
            lambda w, e: software_system.model.id in e.destinations.ids,
        ]

        # TODO: (Or, TOTHINK?) The code below includes all sources and all
        # destinations of the subject software system. What if we want to
        # exclude a source? Maybe the predicates in `elements` and
        # `relationships` should be ANDed together afterall?
        view_relationships_filter: List[Callable[[Workspace, Relationship], bool]] = [
            lambda w, r: software_system == r.source,
            lambda w, r: software_system == r.destination,
        ]

        expression = Expression(
            include_elements=self._include_elements + view_elements_filter,
            exclude_elements=self._exclude_elements,
            include_relationships=self._include_relationships + view_relationships_filter,
            exclude_relationships=self._exclude_relationships,
        )

        workspace = self._parent._parent

        element_ids = map(
            lambda x: str(x.model.id),
            expression.elements(workspace)
        )

        relationship_ids = map(
            lambda x: str(x.model.id),
            expression.relationships(workspace)
        )

        self._m.elements = []
        for element_id in element_ids:
            self._m.elements.append(ElementView(id=element_id))

        self._m.relationships = []
        for relationship_id in relationship_ids:
            self._m.relationships.append(RelationshipView(id=relationship_id))

class ContainerView:

    from buildzr.dsl.expression import Expression, Element, Relationship

    @property
    def model(self) -> buildzr.models.ContainerView:
        return self._m

    @property
    def parent(self) -> Optional['Views']:
        return self._parent

    def __init__(
        self,
        software_system_selector: Callable[[Workspace], SoftwareSystem],
        key: str,
        description: str,
        auto_layout: _AutoLayout='tb',
        title: Optional[str]=None,
        include_elements: List[Callable[[Workspace, Element], bool]]=[],
        exclude_elements: List[Callable[[Workspace, Element], bool]]=[],
        include_relationships: List[Callable[[Workspace, Relationship], bool]]=[],
        exclude_relationships: List[Callable[[Workspace, Relationship], bool]]=[],
        properties: Optional[Dict[str, str]]=None,
    ) -> None:
        self._m = buildzr.models.ContainerView()
        self._parent: Optional['Views'] = None

        self._m.key = key
        self._m.description = description

        self._m.automaticLayout = _auto_layout_to_model(auto_layout)
        self._m.title = title
        self._m.properties = properties

        self._selector = software_system_selector
        self._include_elements = include_elements
        self._exclude_elements = exclude_elements
        self._include_relationships = include_relationships
        self._exclude_relationships = exclude_relationships

    def _on_added(self) -> None:

        from buildzr.dsl.expression import Expression, Element, Relationship
        from buildzr.models import ElementView, RelationshipView

        software_system = self._selector(self._parent._parent)
        self._m.softwareSystemId = software_system.model.id

        container_ids = { container.model.id for container in software_system.children}

        view_elements_filter: List[Callable[[Workspace, Element], bool]] = [
            lambda w, e: e.parent == software_system,
            lambda w, e: any(container_ids.intersection({ id for id in e.sources.ids })),
            lambda w, e: any(container_ids.intersection({ id for id in e.destinations.ids })),
        ]

        view_relationships_filter: List[Callable[[Workspace, Relationship], bool]] = [
            lambda w, r: software_system == r.source.parent,
            lambda w, r: software_system == r.destination.parent,
        ]

        expression = Expression(
            include_elements=self._include_elements + view_elements_filter,
            exclude_elements=self._exclude_elements,
            include_relationships=self._include_relationships + view_relationships_filter,
            exclude_relationships=self._exclude_relationships,
        )

        workspace = self._parent._parent

        element_ids = map(
            lambda x: str(x.model.id),
            expression.elements(workspace)
        )

        relationship_ids = map(
            lambda x: str(x.model.id),
            expression.relationships(workspace)
        )

        self._m.elements = []
        for element_id in element_ids:
            self._m.elements.append(ElementView(id=element_id))

        self._m.relationships = []
        for relationship_id in relationship_ids:
            self._m.relationships.append(RelationshipView(id=relationship_id))

class ComponentView:

    from buildzr.dsl.expression import Expression, Element, Relationship

    @property
    def model(self) -> buildzr.models.ComponentView:
        return self._m

    @property
    def parent(self) -> Optional['Views']:
        return self._parent

    def __init__(
        self,
        container_selector: Callable[[Workspace], Container],
        key: str,
        description: str,
        auto_layout: _AutoLayout='tb',
        title: Optional[str]=None,
        include_elements: List[Callable[[Workspace, Element], bool]]=[],
        exclude_elements: List[Callable[[Workspace, Element], bool]]=[],
        include_relationships: List[Callable[[Workspace, Relationship], bool]]=[],
        exclude_relationships: List[Callable[[Workspace, Relationship], bool]]=[],
        properties: Optional[Dict[str, str]]=None,
    ) -> None:
        self._m = buildzr.models.ComponentView()
        self._parent: Optional['Views'] = None

        self._m.key = key
        self._m.description = description

        self._m.automaticLayout = _auto_layout_to_model(auto_layout)
        self._m.title = title
        self._m.properties = properties

        self._selector = container_selector
        self._include_elements = include_elements
        self._exclude_elements = exclude_elements
        self._include_relationships = include_relationships
        self._exclude_relationships = exclude_relationships

    def _on_added(self) -> None:

        from buildzr.dsl.expression import Expression, Element, Relationship
        from buildzr.models import ElementView, RelationshipView

        container = self._selector(self._parent._parent)
        self._m.containerId = container.model.id

        component_ids = { component.model.id for component in container.children }

        view_elements_filter: List[Callable[[Workspace, Element], bool]] = [
            lambda w, e: e.parent == container,
            lambda w, e: any(component_ids.intersection({ id for id in e.sources.ids })),
            lambda w, e: any(component_ids.intersection({ id for id in e.destinations.ids })),
        ]

        view_relationships_filter: List[Callable[[Workspace, Relationship], bool]] = [
            lambda w, r: container == r.source.parent,
            lambda w, r: container == r.destination.parent,
        ]

        expression = Expression(
            include_elements=self._include_elements + view_elements_filter,
            exclude_elements=self._exclude_elements,
            include_relationships=self._include_relationships + view_relationships_filter,
            exclude_relationships=self._exclude_relationships,
        )

        workspace = self._parent._parent

        element_ids = map(
            lambda x: str(x.model.id),
            expression.elements(workspace)
        )

        relationship_ids = map(
            lambda x: str(x.model.id),
            expression.relationships(workspace)
        )

        self._m.elements = []
        for element_id in element_ids:
            self._m.elements.append(ElementView(id=element_id))

        self._m.relationships = []
        for relationship_id in relationship_ids:
            self._m.relationships.append(RelationshipView(id=relationship_id))

class Views(DslViewsElement):

    @property
    def model(self) -> buildzr.models.Views:
        return self._m

    @property
    def parent(self) -> Optional[Workspace]:
        return self._parent

    def __init__(
        self,
        workspace: Workspace,
    ) -> None:
        self._m = buildzr.models.Views()
        self._parent = workspace
        self._parent._m.views = self._m

    def contains(
        self,
        *views: Union[
            SystemLandscapeView,
            SystemContextView,
            ContainerView,
            ComponentView,
        ]) -> Self:

        for view in views:
            view._parent = self
            if isinstance(view, SystemLandscapeView):
                view._on_added()
                if self._m.systemLandscapeViews:
                    self._m.systemLandscapeViews.append(view.model)
                else:
                    self._m.systemLandscapeViews = [view.model]
            elif isinstance(view, SystemContextView):
                view._on_added()
                if self._m.systemContextViews:
                    self._m.systemContextViews.append(view.model)
                else:
                    self._m.systemContextViews = [view.model]
            elif isinstance(view, ContainerView):
                view._on_added()
                if self._m.containerViews:
                    self._m.containerViews.append(view.model)
                else:
                    self._m.containerViews = [view.model]
            elif isinstance(view, ComponentView):
                view._on_added()
                if self._m.componentViews:
                    self._m.componentViews.append(view.model)
                else:
                    self._m.componentViews = [view.model]
            else:
                raise NotImplementedError("The view {0} is currently not supported", type(view))

        return self

    def get_workspace(self) -> Workspace:
        """
        Get the `Workspace` which contain this views definition.
        """
        return self._parent


from .interfaces import (
    DslElement,
    DslFluentRelationship,
    DslRelationship,
    DslWorkspaceElement,
    DslViewsElement,
    BindLeft,
    BindRight,
    BindLeftLate,
    TSrc,
    TDst,
    TParent,
    TChild,
)